import os
import random
import sys
from typing import NamedTuple, Optional
from unittest.mock import Mock

import numpy as np
import pytest
import qtpy
from qtpy.QtCore import Qt
from qtpy.QtWidgets import (
    QAbstractButton,
    QAbstractSlider,
    QAbstractSpinBox,
    QCheckBox,
    QComboBox,
    QMessageBox,
    QPushButton,
    QRadioButton,
)

from napari._qt.layer_controls.qt_image_controls import QtImageControls
from napari._qt.layer_controls.qt_labels_controls import QtLabelsControls
from napari._qt.layer_controls.qt_layer_controls_container import (
    QtLayerControlsContainer,
    create_qt_layer_controls,
    layer_to_controls,
)
from napari._qt.layer_controls.qt_points_controls import QtPointsControls
from napari._qt.layer_controls.qt_shapes_controls import QtShapesControls
from napari._qt.layer_controls.qt_surface_controls import QtSurfaceControls
from napari._qt.layer_controls.qt_tracks_controls import QtTracksControls
from napari._qt.layer_controls.qt_vectors_controls import QtVectorsControls
from napari._qt.widgets.qt_color_swatch import QColorSwatchEdit
from napari.components import ViewerModel
from napari.layers import (
    Image,
    Labels,
    Layer,
    Points,
    Shapes,
    Surface,
    Tracks,
    Vectors,
)
from napari.utils.colormaps import DirectLabelColormap
from napari.utils.events.event import Event


class LayerTypeWithData(NamedTuple):
    type: type[Layer]
    data: np.ndarray
    colormap: Optional[DirectLabelColormap]
    properties: Optional[dict]
    expected_isinstance: type[QtLayerControlsContainer]


np.random.seed(0)


_IMAGE = LayerTypeWithData(
    type=Image,
    data=np.random.rand(8, 8),
    colormap=None,
    properties=None,
    expected_isinstance=QtImageControls,
)
_LABELS_WITH_DIRECT_COLORMAP = LayerTypeWithData(
    type=Labels,
    data=np.random.randint(5, size=(10, 15)),
    colormap=DirectLabelColormap(
        color_dict={
            1: 'white',
            2: 'blue',
            3: 'green',
            4: 'red',
            5: 'yellow',
            None: 'black',
        }
    ),
    properties=None,
    expected_isinstance=QtLabelsControls,
)
_LABELS = LayerTypeWithData(
    type=Labels,
    data=np.random.randint(5, size=(10, 15)),
    colormap=None,
    properties=None,
    expected_isinstance=QtLabelsControls,
)
_POINTS = LayerTypeWithData(
    type=Points,
    data=np.random.random((5, 2)),
    colormap=None,
    properties=None,
    expected_isinstance=QtPointsControls,
)
_SHAPES = LayerTypeWithData(
    type=Shapes,
    data=np.random.random((10, 4, 2)),
    colormap=None,
    properties=None,
    expected_isinstance=QtShapesControls,
)
_SURFACE = LayerTypeWithData(
    type=Surface,
    data=(
        np.random.random((10, 2)),
        np.random.randint(10, size=(6, 3)),
        np.random.random(10),
    ),
    colormap=None,
    properties=None,
    expected_isinstance=QtSurfaceControls,
)
_TRACKS = LayerTypeWithData(
    type=Tracks,
    data=np.zeros((2, 4)),
    colormap=None,
    properties={
        'track_id': [0, 0],
        'time': [0, 0],
        'speed': [50, 30],
    },
    expected_isinstance=QtTracksControls,
)
_VECTORS = LayerTypeWithData(
    type=Vectors,
    data=np.zeros((2, 2, 2)),
    colormap=None,
    properties=None,
    expected_isinstance=QtVectorsControls,
)
_LINES_DATA = np.random.random((6, 2, 2))


@pytest.fixture
def create_layer_controls(qtbot):
    def _create_layer_controls(layer_type_with_data):
        if layer_type_with_data.colormap:
            layer = layer_type_with_data.type(
                layer_type_with_data.data,
                colormap=layer_type_with_data.colormap,
            )
        elif layer_type_with_data.properties:
            layer = layer_type_with_data.type(
                layer_type_with_data.data,
                properties=layer_type_with_data.properties,
            )
        else:
            layer = layer_type_with_data.type(layer_type_with_data.data)

        ctrl = create_qt_layer_controls(layer)
        qtbot.addWidget(ctrl)

        return ctrl

    return _create_layer_controls


@pytest.mark.parametrize(
    'layer_type_with_data',
    [
        _LABELS_WITH_DIRECT_COLORMAP,
        _LABELS,
        _IMAGE,
        _POINTS,
        _SHAPES,
        _SURFACE,
        _TRACKS,
        _VECTORS,
    ],
    ids=[
        'labels_with_direct_colormap',
        'labels_with_auto_colormap',
        'image',
        'points',
        'shapes',
        'surface',
        'tracks',
        'vectors',
    ],
)
@pytest.mark.qt_no_exception_capture
@pytest.mark.skipif(os.environ.get('MIN_REQ', '0') == '1', reason='min req')
def test_create_layer_controls(
    qtbot, create_layer_controls, layer_type_with_data, capsys
):
    # create layer controls widget
    ctrl = create_layer_controls(layer_type_with_data)

    # check create widget corresponds to the expected class for each type of layer
    assert isinstance(ctrl, layer_type_with_data.expected_isinstance)

    # check QComboBox by changing current index
    for qcombobox in ctrl.findChildren(QComboBox):
        if qcombobox.isVisible():
            qcombobox_count = qcombobox.count()
            qcombobox_initial_idx = qcombobox.currentIndex()
            if qcombobox_count:
                qcombobox.setCurrentIndex(0)
            for idx in range(qcombobox_count):
                previous_qcombobox_text = qcombobox.currentText()
                qcombobox.setCurrentIndex(idx)
                # If a value for the QComboBox is an invalid selection check if
                # it fallbacks to the previous value
                captured = capsys.readouterr()
                if captured.err:
                    assert qcombobox.currentText() == previous_qcombobox_text
            qcombobox.setCurrentIndex(qcombobox_initial_idx)


skip_predicate = sys.version_info >= (3, 11) and (
    qtpy.API == 'pyqt5' or qtpy.API == 'pyqt6'
)


@pytest.mark.parametrize(
    'layer_type_with_data',
    [
        # those 2 fail on 3.11 + pyqt5 and pyqt6 with a segfault that can't be caught by
        # pytest in qspinbox.setValue(value)
        # See: https://github.com/napari/napari/pull/5439
        pytest.param(
            _LABELS_WITH_DIRECT_COLORMAP,
            marks=pytest.mark.skipif(
                skip_predicate,
                reason='segfault on Python 3.11+ and pyqt5 or Pyqt6',
            ),
        ),
        pytest.param(
            _LABELS,
            marks=pytest.mark.skipif(
                skip_predicate,
                reason='segfault on Python 3.11+ and pyqt5 or Pyqt6',
            ),
        ),
        _IMAGE,
        _POINTS,
        _SHAPES,
        _SURFACE,
        _TRACKS,
        _VECTORS,
    ],
)
@pytest.mark.qt_no_exception_capture
@pytest.mark.skipif(os.environ.get('MIN_REQ', '0') == '1', reason='min req')
def test_create_layer_controls_spin(
    qtbot, create_layer_controls, layer_type_with_data, capsys
):
    # create layer controls widget
    ctrl = create_layer_controls(layer_type_with_data)
    qtbot.addWidget(ctrl)

    # check create widget corresponds to the expected class for each type of layer
    assert isinstance(ctrl, layer_type_with_data.expected_isinstance)

    # check QAbstractSpinBox by changing value with `setValue` from minimum value to maximum
    for qspinbox in ctrl.findChildren(QAbstractSpinBox):
        qspinbox_initial_value = qspinbox.value()
        qspinbox_min = qspinbox.minimum()
        qspinbox_max = qspinbox.maximum()
        if isinstance(qspinbox_min, float):
            if np.isinf(qspinbox_max):
                qspinbox_max = sys.float_info.max
            value_range = np.linspace(qspinbox_min, qspinbox_max)
        else:
            # use + 1 to include maximum value
            value_range = range(qspinbox_min, qspinbox_max + 1)
            try:
                value_range_length = len(value_range)
            except OverflowError:
                # range too big for even trying to get how big it is.
                value_range_length = 100
                value_range = [
                    random.randrange(qspinbox_min, qspinbox_max)
                    for _ in range(value_range_length)
                ]
                value_range.append(qspinbox_max)
            if value_range_length > 100:
                # prevent iterating over a big range of values
                random.seed(0)
                value_range = random.sample(value_range, 100)
                value_range = np.insert(value_range, 0, qspinbox_min)
                value_range = np.append(value_range, qspinbox_max - 1)
        for value in value_range:
            qspinbox.setValue(value)
            # capture any output done to sys.stdout or sys.stderr.
            captured = capsys.readouterr()
            assert not captured.out
            if captured.err:
                # since an error was found check if it is associated with a known issue still open
                expected_errors = [
                    'MemoryError: Unable to allocate',  # See https://github.com/napari/napari/issues/5798
                    'ValueError: array is too big; `arr.size * arr.dtype.itemsize` is larger than the maximum possible size.',  # See https://github.com/napari/napari/issues/5798
                    'ValueError: Maximum allowed dimension exceeded',  # See https://github.com/napari/napari/issues/5798
                    'IndexError: index ',  # See https://github.com/napari/napari/issues/4864
                    'RuntimeWarning: overflow encountered',  # See https://github.com/napari/napari/issues/4864
                ]
                assert any(
                    expected_error in captured.err
                    for expected_error in expected_errors
                ), f'value: {value}, range {value_range}\nerr: {captured.err}'

        assert qspinbox.value() in [qspinbox_max, qspinbox_max - 1]
        qspinbox.setValue(qspinbox_initial_value)


@pytest.mark.parametrize(
    'layer_type_with_data',
    [
        _LABELS_WITH_DIRECT_COLORMAP,
        _LABELS,
        _IMAGE,
        _POINTS,
        _SHAPES,
        _SURFACE,
        _TRACKS,
        _VECTORS,
    ],
)
@pytest.mark.qt_no_exception_capture
@pytest.mark.skipif(os.environ.get('MIN_REQ', '0') == '1', reason='min req')
def test_create_layer_controls_qslider(
    qtbot, create_layer_controls, layer_type_with_data, capsys
):
    # create layer controls widget
    ctrl = create_layer_controls(layer_type_with_data)

    # check create widget corresponds to the expected class for each type of layer
    assert isinstance(ctrl, layer_type_with_data.expected_isinstance)

    # check QAbstractSlider by changing value with `setValue` from minimum value to maximum
    for qslider in ctrl.findChildren(QAbstractSlider):
        if isinstance(qslider.minimum(), float):
            if getattr(qslider, '_valuesChanged', None):
                # create a list of tuples in the case the slider is ranged
                # from (minimum, minimum) to (maximum, maximum) +
                # from (minimum, maximum) to (minimum, minimum)
                # (minimum, minimum) and (maximum, maximum) values are excluded
                # to prevent the sequence not being monotonically increasing
                base_value_range = np.linspace(
                    qslider.minimum(), qslider.maximum()
                )
                num_values = base_value_range.size
                max_value = np.full(num_values, qslider.maximum())
                min_value = np.full(num_values, qslider.minimum())
                value_range_to_max = list(zip(base_value_range, max_value))
                value_range_to_min = list(
                    zip(min_value, np.flip(base_value_range))
                )
                value_range = value_range_to_max[:-1] + value_range_to_min[:-1]
            else:
                value_range = np.linspace(qslider.minimum(), qslider.maximum())
        else:
            if getattr(qslider, '_valuesChanged', None):
                # create a list of tuples in the case the slider is ranged
                # from (minimum, minimum) to (maximum, maximum) +
                # from (minimum, maximum) to (minimum, minimum)
                # base list created with + 1 to include maximum value
                # (minimum, minimum) and (maximum, maximum) values are excluded
                # to prevent the sequence not being monotonically increasing
                base_value_range = range(
                    qslider.minimum(), qslider.maximum() + 1
                )
                num_values = len(base_value_range)
                max_value = [qslider.maximum()] * num_values
                min_value = [qslider.minimum()] * num_values
                value_range_to_max = list(zip(base_value_range, max_value))
                base_value_range_copy = base_value_range.copy()
                base_value_range_copy.reverse()
                value_range_to_min = list(
                    zip(min_value, base_value_range_copy)
                )
                value_range = value_range_to_max[:-1] + value_range_to_min[:-1]
            else:
                # use + 1 to include maximum value
                value_range = range(qslider.minimum(), qslider.maximum() + 1)
        for value in value_range:
            qslider.setValue(value)
            # capture any output done to sys.stdout or sys.stderr.
            captured = capsys.readouterr()
            assert not captured.out
            assert not captured.err
        if getattr(qslider, '_valuesChanged', None):
            assert qslider.value()[0] == qslider.minimum()
        else:
            assert qslider.value() == qslider.maximum()


@pytest.mark.parametrize(
    'layer_type_with_data',
    [
        _LABELS_WITH_DIRECT_COLORMAP,
        _LABELS,
        _IMAGE,
        _POINTS,
        _SHAPES,
        _SURFACE,
        _TRACKS,
        _VECTORS,
    ],
)
@pytest.mark.qt_no_exception_capture
@pytest.mark.skipif(os.environ.get('MIN_REQ', '0') == '1', reason='min req')
def test_create_layer_controls_qcolorswatchedit(
    qtbot, create_layer_controls, layer_type_with_data, capsys
):
    # create layer controls widget
    ctrl = create_layer_controls(layer_type_with_data)

    # check create widget corresponds to the expected class for each type of layer
    assert isinstance(ctrl, layer_type_with_data.expected_isinstance)

    # check QColorSwatchEdit by changing line edit text with a range of predefined values
    for qcolorswatchedit in ctrl.findChildren(QColorSwatchEdit):
        lineedit = qcolorswatchedit.line_edit
        colorswatch = qcolorswatchedit.color_swatch
        colors = [
            ('white', 'white', np.array([1.0, 1.0, 1.0, 1.0])),
            ('black', 'black', np.array([0.0, 0.0, 0.0, 1.0])),
            # check autocompletion `bla` -> `black`
            ('bla', 'black', np.array([0.0, 0.0, 0.0, 1.0])),
            # check that setting an invalid color makes it fallback to the previous value
            ('invalid_value', 'black', np.array([0.0, 0.0, 0.0, 1.0])),
        ]
        for color, expected_color, expected_array in colors:
            lineedit.clear()
            qtbot.keyClicks(lineedit, color)
            qtbot.keyClick(lineedit, Qt.Key_Enter)
            assert lineedit.text() == expected_color
            assert (colorswatch.color == expected_array).all()
            # capture any output done to sys.stdout or sys.stderr.
            captured = capsys.readouterr()
            assert not captured.out
            assert not captured.err

    # check QCheckBox by clicking with mouse click
    for qcheckbox in ctrl.findChildren(QCheckBox):
        if qcheckbox.isVisible():
            qcheckbox_checked = qcheckbox.isChecked()
            qtbot.mouseClick(qcheckbox, Qt.LeftButton)
            assert qcheckbox.isChecked() != qcheckbox_checked
            # capture any output done to sys.stdout or sys.stderr.
            captured = capsys.readouterr()
            assert not captured.out
            assert not captured.err

    # check QPushButton and QRadioButton by clicking with mouse click
    for button in ctrl.findChildren(QPushButton) + ctrl.findChildren(
        QRadioButton
    ):
        if button.isVisible():
            qtbot.mouseClick(button, Qt.LeftButton)
            # capture any output done to sys.stdout or sys.stderr.
            captured = capsys.readouterr()
            assert not captured.out
            assert not captured.err


@pytest.mark.parametrize(
    (
        'layer_type_with_data',
        'action_manager_trigger',
    ),
    [
        (
            _LABELS_WITH_DIRECT_COLORMAP,
            'napari:activate_labels_transform_mode',
        ),
        (
            _LABELS,
            'napari:activate_labels_transform_mode',
        ),
        (
            _IMAGE,
            'napari:activate_image_transform_mode',
        ),
        (
            _POINTS,
            'napari:activate_points_transform_mode',
        ),
        (
            _SHAPES,
            'napari:activate_shapes_transform_mode',
        ),
        (
            _SURFACE,
            'napari:activate_surface_transform_mode',
        ),
        (
            _TRACKS,
            'napari:activate_tracks_transform_mode',
        ),
        (
            _VECTORS,
            'napari:activate_vectors_transform_mode',
        ),
    ],
)
def test_create_layer_controls_transform_mode_button(
    qtbot,
    create_layer_controls,
    layer_type_with_data,
    action_manager_trigger,
    monkeypatch,
):
    action_manager_mock = Mock(trigger=Mock())

    # Monkeypatch the action_manager instance to prevent `KeyError: 'layer'`
    # over `napari.layers.utils.layer_utils.register_layer_attr_action._handle._wrapper`
    monkeypatch.setattr(
        'napari._qt.layer_controls.qt_layer_controls_base.action_manager',
        action_manager_mock,
    )

    # create layer controls widget
    ctrl = create_layer_controls(layer_type_with_data)

    # check create widget corresponds to the expected class for each type of layer
    assert isinstance(ctrl, layer_type_with_data.expected_isinstance)

    # check transform mode button existence
    assert ctrl.transform_button

    # check layer mode change
    assert ctrl.layer.mode == 'pan_zoom'
    ctrl.transform_button.click()
    assert ctrl.layer.mode == 'transform'

    # check reset transform behavior
    ctrl.layer.affine = None
    assert ctrl.layer.affine != ctrl.layer._initial_affine

    def reset_transform_warning_dialog(*args):
        return QMessageBox.Yes

    monkeypatch.setattr(
        'qtpy.QtWidgets.QMessageBox.warning', reset_transform_warning_dialog
    )
    qtbot.mouseClick(
        ctrl.transform_button,
        Qt.LeftButton,
        Qt.KeyboardModifier.AltModifier,
    )
    assert ctrl.layer.affine == ctrl.layer._initial_affine


@pytest.mark.parametrize(
    'layer_type_with_data',
    [
        _LABELS_WITH_DIRECT_COLORMAP,
        _LABELS,
        _IMAGE,
        _POINTS,
        _SHAPES,
        _SURFACE,
        _TRACKS,
        _VECTORS,
    ],
)
def test_layer_controls_invalid_mode(
    qtbot,
    create_layer_controls,
    layer_type_with_data,
):
    # create layer controls widget
    ctrl = create_layer_controls(layer_type_with_data)

    # check create widget corresponds to the expected class for each type of layer
    assert isinstance(ctrl, layer_type_with_data.expected_isinstance)

    # check layer mode and corresponding mode button
    assert ctrl.layer.mode == 'pan_zoom'
    assert ctrl.panzoom_button.isChecked()

    # check setting invalid mode
    with pytest.raises(ValueError, match='not recognized'):
        ctrl._on_mode_change(Event('mode', mode='invalid_mode'))

    # check panzoom_button is still checked
    assert ctrl.panzoom_button.isChecked()


def test_unknown_raises(qtbot):
    class Test:
        """Unmatched class"""

    with pytest.raises(TypeError):
        create_qt_layer_controls(Test())


def test_inheritance(qtbot):
    class QtLinesControls(QtShapesControls):
        """Yes I'm the same"""

    class Lines(Shapes):
        """Here too"""

    lines = Lines(_LINES_DATA)
    layer_to_controls[Lines] = QtLinesControls
    ctrl = create_qt_layer_controls(lines)
    qtbot.addWidget(ctrl)
    assert isinstance(ctrl, QtLinesControls)


@pytest.mark.parametrize('layer_type_with_data', [_POINTS, _SHAPES])
def test_text_set_visible_updates_checkbox(qtbot, layer_type_with_data):
    text = {
        'string': {'constant': 'test'},
        'visible': True,
    }
    layer = layer_type_with_data.type(layer_type_with_data.data, text=text)
    ctrl = create_qt_layer_controls(layer)
    qtbot.addWidget(ctrl)
    assert ctrl.textDispCheckBox.isChecked()

    layer.text.visible = False

    assert not ctrl.textDispCheckBox.isChecked()


@pytest.mark.parametrize('layer_type_with_data', [_POINTS, _SHAPES])
def test_set_text_then_set_visible_updates_checkbox(
    qtbot, layer_type_with_data
):
    layer = layer_type_with_data.type(layer_type_with_data.data)
    ctrl = create_qt_layer_controls(layer)
    qtbot.addWidget(ctrl)
    layer.text = {
        'string': {'constant': 'another_test'},
        'visible': False,
    }
    assert not ctrl.textDispCheckBox.isChecked()

    layer.text.visible = True

    assert ctrl.textDispCheckBox.isChecked()


@pytest.mark.parametrize(('ndim', 'editable_after'), [(2, False), (3, True)])
def test_set_3d_display_with_points(qtbot, ndim, editable_after):
    """Interactivity only works for 2D points layers rendered in 2D and not
    in 3D. Verify that layer.editable is set appropriately upon switching to
    3D rendering mode.

    See: https://github.com/napari/napari/pull/4184
    """
    viewer = ViewerModel()
    container = QtLayerControlsContainer(viewer)
    qtbot.addWidget(container)
    layer = viewer.add_points(np.zeros((0, ndim)), ndim=ndim)
    assert viewer.dims.ndisplay == 2
    assert layer.editable

    viewer.dims.ndisplay = 3

    assert layer.editable == editable_after


def test_set_3d_display_with_shapes(qtbot):
    """Interactivity only works for shapes layers rendered in 2D and not
    in 3D. Verify that layer.editable is set appropriately upon switching to
    3D rendering mode.

    See: https://github.com/napari/napari/pull/4184
    """
    viewer = ViewerModel()
    container = QtLayerControlsContainer(viewer)
    qtbot.addWidget(container)
    layer = viewer.add_shapes(np.zeros((0, 2, 4)))
    assert viewer.dims.ndisplay == 2
    assert layer.editable

    viewer.dims.ndisplay = 3

    assert not layer.editable


def test_set_3d_display_with_labels(qtbot):
    """Some modes only work for labels layers rendered in 2D and not
    in 3D. Verify that the related mode buttons are disabled upon switching to
    3D rendering mode while the layer is still editable.
    """
    viewer = ViewerModel()
    container = QtLayerControlsContainer(viewer)
    qtbot.addWidget(container)
    layer = viewer.add_labels(np.zeros((3, 4), dtype=int))
    assert viewer.dims.ndisplay == 2
    assert container.currentWidget().polygon_button.isEnabled()
    assert container.currentWidget().transform_button.isEnabled()
    assert layer.editable

    viewer.dims.ndisplay = 3

    assert not container.currentWidget().polygon_button.isEnabled()
    assert not container.currentWidget().transform_button.isEnabled()
    assert layer.editable


@pytest.mark.parametrize(
    'add_layer_with_data',
    [
        ('add_labels', np.zeros((3, 4), dtype=int)),
        ('add_points', np.empty((0, 2))),
        ('add_shapes', np.empty((0, 2, 4))),
        ('add_image', np.random.rand(8, 8)),
        (
            'add_surface',
            (
                np.random.random((10, 2)),
                np.random.randint(10, size=(6, 3)),
                np.random.random(10),
            ),
        ),
        ('add_tracks', np.zeros((2, 4))),
        ('add_vectors', np.zeros((2, 2, 2))),
    ],
)
def test_set_3d_display_and_layer_visibility(qtbot, add_layer_with_data):
    """Some modes only work for layers rendered in 2D and not
    in 3D. Verify that the related mode buttons are disabled upon switching to
    3D rendering mode and the disable state is kept even when changing layer
    visibility.

    For the labels layer the specific polygon mode button should be disabled in
    3D regardless of the layer being visible or not. For all the layers the same
    applies for the transform mode button.
    """
    viewer = ViewerModel()
    container = QtLayerControlsContainer(viewer)
    qtbot.addWidget(container)
    add_layer_method, data = add_layer_with_data
    layer = getattr(viewer, add_layer_method)(data)

    # 2D mode
    assert viewer.dims.ndisplay == 2
    if add_layer_method == 'add_labels':
        assert container.currentWidget().polygon_button.isEnabled()
    assert container.currentWidget().transform_button.isEnabled()

    # 2D mode + layer not visible
    layer.visible = False
    if add_layer_method == 'add_labels':
        assert not container.currentWidget().polygon_button.isEnabled()
    assert not container.currentWidget().transform_button.isEnabled()

    # 2D mode + layer visible
    layer.visible = True
    if add_layer_method == 'add_labels':
        assert container.currentWidget().polygon_button.isEnabled()
    assert container.currentWidget().transform_button.isEnabled()

    # 3D mode
    viewer.dims.ndisplay = 3
    if add_layer_method == 'add_labels':
        assert not container.currentWidget().polygon_button.isEnabled()
    assert not container.currentWidget().transform_button.isEnabled()

    # 3D mode + layer not visible
    layer.visible = False
    if add_layer_method == 'add_labels':
        assert not container.currentWidget().polygon_button.isEnabled()
    assert not container.currentWidget().transform_button.isEnabled()

    # 3D mode + layer visible
    layer.visible = True
    if add_layer_method == 'add_labels':
        assert not container.currentWidget().polygon_button.isEnabled()
    assert not container.currentWidget().transform_button.isEnabled()


# The following tests handle changes to the layer's visible and
# editable state for layer control types that have controls to edit
# the layer. For more context see:
# https://github.com/napari/napari/issues/1346
# Updated due to the addition of a transform mode button for all the layers,
# For more context see:
# https://github.com/napari/napari/pull/6794


@pytest.fixture(
    params=(
        (Labels, np.zeros((3, 4), dtype=int)),
        (Points, np.empty((0, 2))),
        (Shapes, np.empty((0, 2, 4))),
        (Image, np.random.rand(8, 8)),
        (
            Surface,
            (
                np.random.random((10, 2)),
                np.random.randint(10, size=(6, 3)),
                np.random.random(10),
            ),
        ),
        (Tracks, np.zeros((2, 4))),
        (Vectors, np.zeros((2, 2, 2))),
    )
)
def editable_layer(request):
    LayerType, data = request.param
    return LayerType(data)


def test_make_visible_when_editable_enables_edit_buttons(
    qtbot, editable_layer
):
    editable_layer.editable = True
    editable_layer.visible = False
    controls = make_layer_controls(qtbot, editable_layer)
    assert_no_edit_buttons_enabled(controls)

    editable_layer.visible = True

    assert_all_edit_buttons_enabled(controls)


def test_make_not_visible_when_editable_disables_edit_buttons(
    qtbot, editable_layer
):
    editable_layer.editable = True
    editable_layer.visible = True
    controls = make_layer_controls(qtbot, editable_layer)
    assert_all_edit_buttons_enabled(controls)

    editable_layer.visible = False

    assert_no_edit_buttons_enabled(controls)


def test_make_editable_when_visible_enables_edit_buttons(
    qtbot, editable_layer
):
    editable_layer.editable = False
    editable_layer.visible = True
    controls = make_layer_controls(qtbot, editable_layer)
    assert_no_edit_buttons_enabled(controls)

    editable_layer.editable = True

    assert_all_edit_buttons_enabled(controls)


def test_make_not_editable_when_visible_disables_edit_buttons(
    qtbot, editable_layer
):
    editable_layer.editable = True
    editable_layer.visible = True
    controls = make_layer_controls(qtbot, editable_layer)
    assert_all_edit_buttons_enabled(controls)

    editable_layer.editable = False

    assert_no_edit_buttons_enabled(controls)


def make_layer_controls(qtbot, layer):
    QtLayerControlsType = layer_to_controls[type(layer)]
    controls = QtLayerControlsType(layer)
    qtbot.addWidget(controls)
    return controls


def assert_all_edit_buttons_enabled(controls) -> None:
    assert all(map(QAbstractButton.isEnabled, controls._EDIT_BUTTONS))


def assert_no_edit_buttons_enabled(controls) -> None:
    assert not any(map(QAbstractButton.isEnabled, controls._EDIT_BUTTONS))



"""MutableMapping that emits events when altered."""

from collections.abc import Mapping, Sequence
from typing import Optional, Union

from napari.utils.events.containers._dict import _K, _T, TypedMutableMapping
from napari.utils.events.event import EmitterGroup, Event
from napari.utils.events.types import SupportsEvents


class EventedDict(TypedMutableMapping[_K, _T]):
    """Mutable dictionary that emits events when altered.

    This class is designed to behave exactly like builtin ``dict``, but
    will emit events before and after all mutations (addition, removal, and
    changing).

    Parameters
    ----------
    data : Mapping, optional
        Dictionary to initialize the class with.
    basetype : type of sequence of types, optional
        Type of the element in the dictionary.

    Events
    ------
    changing (key: K)
        emitted before an item at ``key`` is changed
    changed (key: K, old_value: T, value: T)
        emitted when item at ``key`` is changed from ``old_value`` to ``value``
    adding (key: K)
        emitted before an item is added to the dictionary with ``key``
    added (key: K, value: T)
        emitted after ``value`` was added to the dictionary with ``key``
    removing (key: K)
        emitted before ``key`` is removed from the dictionary
    removed (key: K, value: T)
        emitted after ``key`` was removed from the dictionary
    updated (key, K, value: T)
        emitted after ``value`` of ``key`` was changed. Only implemented by
        subclasses to give them an option to trigger some update after ``value``
        was changed and this class did not register it. This can be useful if
        the ``basetype`` is not an evented object.
    """

    events: EmitterGroup

    def __init__(
        self,
        data: Optional[Mapping[_K, _T]] = None,
        basetype: Union[type[_T], Sequence[type[_T]]] = (),
    ) -> None:
        _events = {
            'changing': None,
            'changed': None,
            'adding': None,
            'added': None,
            'removing': None,
            'removed': None,
            'updated': None,
        }
        # For inheritance: If the mro already provides an EmitterGroup, add...
        if hasattr(self, 'events') and isinstance(self.events, EmitterGroup):
            self.events.add(**_events)
        else:
            # otherwise create a new one
            self.events = EmitterGroup(
                source=self, auto_connect=False, **_events
            )
        super().__init__(data, basetype)

    def __setitem__(self, key: _K, value: _T) -> None:
        old = self._dict.get(key)
        if value is old or value == old:
            return
        if old is None:
            self.events.adding(key=key)
            super().__setitem__(key, value)
            self.events.added(key=key, value=value)
            self._connect_child_emitters(value)
        else:
            self.events.changing(key=key)
            super().__setitem__(key, value)
            self.events.changed(key=key, old_value=old, value=value)

    def __delitem__(self, key: _K) -> None:
        self.events.removing(key=key)
        self._disconnect_child_emitters(self[key])
        item = self._dict.pop(key)
        self.events.removed(key=key, value=item)

    def _reemit_child_event(self, event: Event) -> None:
        """An item in the dict emitted an event.  Re-emit with key"""
        if not hasattr(event, 'key'):
            event.key = self.key(event.source)

        # re-emit with this object's EventEmitter
        self.events(event)

    def _disconnect_child_emitters(self, child: _T) -> None:
        """Disconnect all events from the child from the re-emitter."""
        if isinstance(child, SupportsEvents):
            child.events.disconnect(self._reemit_child_event)

    def _connect_child_emitters(self, child: _T) -> None:
        """Connect all events from the child to be re-emitted."""
        if isinstance(child, SupportsEvents):
            # make sure the event source has been set on the child
            if child.events.source is None:
                child.events.source = child
            child.events.connect(self._reemit_child_event)

    def key(self, value: _T) -> Optional[_K]:
        """Return first instance of value."""
        for k, v in self._dict.items():
            if v is value or v == value:
                return k
        return None



from napari.utils.transforms.transform_utils import shear_matrix_from_angle
from napari.utils.transforms.transforms import (
    Affine,
    CompositeAffine,
    ScaleTranslate,
    Transform,
    TransformChain,
)

__all__ = [
    'shear_matrix_from_angle',
    'Affine',
    'CompositeAffine',
    'ScaleTranslate',
    'Transform',
    'TransformChain',
]



"""Text label model."""

from napari._pydantic_compat import Field
from napari.components.overlays.base import CanvasOverlay
from napari.utils.color import ColorValue


class TextOverlay(CanvasOverlay):
    """Label model to display arbitrary text in the canvas

    Attributes
    ----------
    color : np.ndarray
        A (4,) color array of the text overlay.
    font_size : float
        The font size (in points) of the text.
    text : str
        Text to be displayed in the canvas.
    position : CanvasPosition
        The position of the overlay in the canvas.
    visible : bool
        If the overlay is visible or not.
    opacity : float
        The opacity of the overlay. 0 is fully transparent.
    order : int
        The rendering order of the overlay: lower numbers get rendered first.
    """

    color: ColorValue = Field(
        default_factory=lambda: ColorValue((0.5, 0.5, 0.5, 1.0))
    )
    font_size: float = 10
    text: str = ''



import json
from enum import EnumMeta
from typing import TYPE_CHECKING, ClassVar, cast

from qtpy.QtCore import QSize, Qt, Signal
from qtpy.QtWidgets import (
    QApplication,
    QDialog,
    QHBoxLayout,
    QListWidget,
    QMessageBox,
    QPushButton,
    QScrollArea,
    QStackedWidget,
    QVBoxLayout,
)

from napari._pydantic_compat import BaseModel, ModelField, ModelMetaclass
from napari.utils.compat import StrEnum
from napari.utils.translations import trans

if TYPE_CHECKING:
    from qtpy.QtGui import QCloseEvent, QKeyEvent


class PreferencesDialog(QDialog):
    """Preferences Dialog for Napari user settings."""

    ui_schema: ClassVar[dict[str, dict[str, str]]] = {
        'call_order': {'ui:widget': 'plugins'},
        'highlight': {'ui:widget': 'highlight'},
        'shortcuts': {'ui:widget': 'shortcuts'},
        'extension2reader': {'ui:widget': 'extension2reader'},
        'dask': {'ui:widget': 'horizontal_object'},
        'font_size': {'ui:widget': 'font_size'},
    }

    resized = Signal(QSize)

    def __init__(self, parent=None) -> None:
        from napari.settings import get_settings

        super().__init__(parent)
        self.setWindowTitle(trans._('Preferences'))
        self.setMinimumSize(QSize(1065, 470))

        self._settings = get_settings()
        self._stack = QStackedWidget(self)
        self._list = QListWidget(self)
        self._list.setObjectName('Preferences')
        self._list.currentRowChanged.connect(self._stack.setCurrentIndex)

        # Set up buttons
        self._button_cancel = QPushButton(trans._('Cancel'))
        self._button_cancel.clicked.connect(self.reject)
        self._button_ok = QPushButton(trans._('OK'))
        self._button_ok.clicked.connect(self.accept)
        self._button_ok.setDefault(True)
        self._button_restore = QPushButton(trans._('Restore defaults'))
        self._button_restore.clicked.connect(self._restore_default_dialog)

        # Layout
        left_layout = QVBoxLayout()
        left_layout.addWidget(self._list)
        left_layout.addStretch()
        left_layout.addWidget(self._button_restore)
        left_layout.addWidget(self._button_cancel)
        left_layout.addWidget(self._button_ok)

        self.setLayout(QHBoxLayout())
        self.layout().addLayout(left_layout, 1)
        self.layout().addWidget(self._stack, 4)

        # Build dialog from settings
        self._rebuild_dialog()

    def keyPressEvent(self, e: 'QKeyEvent'):
        if e.key() == Qt.Key.Key_Escape:
            # escape key should just close the window
            # which implies "accept"
            e.accept()
            self.accept()
            return
        super().keyPressEvent(e)

    def resizeEvent(self, event):
        """Override to emit signal."""
        self.resized.emit(event.size())
        super().resizeEvent(event)

    def _rebuild_dialog(self):
        """Removes settings not to be exposed to user and creates dialog pages."""
        # FIXME: this dialog should not need to know about the plugin manager
        from napari.plugins import plugin_manager

        self._starting_pm_order = plugin_manager.call_order()
        self._starting_values = self._settings.dict(exclude={'schema_version'})

        self._list.clear()
        while self._stack.count():
            self._stack.removeWidget(self._stack.currentWidget())

        for field in self._settings.__fields__.values():
            if isinstance(field.type_, type) and issubclass(
                field.type_, BaseModel
            ):
                self._add_page(field)

        self._list.setCurrentRow(0)

    def _add_page(self, field: 'ModelField'):
        """Builds the preferences widget using the json schema builder.

        Parameters
        ----------
        field : ModelField
            subfield for which to create a page.
        """
        from napari._vendor.qt_json_builder.qt_jsonschema_form import (
            WidgetBuilder,
        )

        schema, values = self._get_page_dict(field)
        name = field.field_info.title or field.name

        form = WidgetBuilder().create_form(schema, self.ui_schema)
        # set state values for widget
        form.widget.state = values
        # make settings follow state of the form widget
        form.widget.on_changed.connect(
            lambda d: getattr(self._settings, name.lower()).update(d)
        )
        # make widgets follow values of the settings
        settings_category = getattr(self._settings, name.lower())
        excluded = set(
            getattr(
                getattr(settings_category, 'NapariConfig', None),
                'preferences_exclude',
                {},
            )
        )
        nested_settings = ['dask', 'highlight']
        for name_, emitter in settings_category.events.emitters.items():
            if name_ not in excluded and name_ not in nested_settings:
                emitter.connect(update_widget_state(name_, form.widget))
            elif name_ in nested_settings:
                # Needed to handle nested event model settings (i.e `DaskSettings` and `HighlightSettings`)
                for subname_, subemitter in getattr(
                    settings_category, name_
                ).events.emitters.items():
                    subemitter.connect(
                        update_widget_state(
                            subname_, form.widget.widgets[name_]
                        )
                    )

        page_scrollarea = QScrollArea()
        page_scrollarea.setWidgetResizable(True)
        page_scrollarea.setWidget(form)

        self._list.addItem(field.field_info.title or field.name)
        self._stack.addWidget(page_scrollarea)

    def _get_page_dict(self, field: 'ModelField') -> tuple[dict, dict]:
        """Provides the schema, set of values for each setting, and the
        properties for each setting."""
        ftype = cast('BaseModel', field.type_)

        # TODO make custom shortcuts dialog to properly capture new
        #      functionality once we switch to app-model's keybinding system
        #      then we can remove the below code used for autogeneration
        if field.name == 'shortcuts':
            # hardcode workaround because pydantic's schema generation
            # does not allow you to specify custom JSON serialization
            schema = {
                'title': 'ShortcutsSettings',
                'type': 'object',
                'properties': {
                    'shortcuts': {
                        'title': field.type_.__fields__[
                            'shortcuts'
                        ].field_info.title,
                        'description': field.type_.__fields__[
                            'shortcuts'
                        ].field_info.description,
                        'type': 'object',
                    }
                },
            }
        else:
            schema = json.loads(ftype.schema_json())

        if field.field_info.title:
            schema['title'] = field.field_info.title
        if field.field_info.description:
            schema['description'] = field.field_info.description

        # find enums:
        for name, subfield in ftype.__fields__.items():
            if isinstance(subfield.type_, EnumMeta):
                enums = [s.value for s in subfield.type_]  # type: ignore
                schema['properties'][name]['enum'] = enums
                schema['properties'][name]['type'] = 'string'
            if isinstance(subfield.type_, ModelMetaclass):
                local_schema = json.loads(subfield.type_.schema_json())
                schema['properties'][name]['type'] = 'object'
                schema['properties'][name]['properties'] = local_schema[
                    'properties'
                ]

        # Need to remove certain properties that will not be displayed on the GUI
        setting = getattr(self._settings, field.name)
        with setting.enums_as_values():
            values = setting.dict()
        napari_config = getattr(setting, 'NapariConfig', None)
        if hasattr(napari_config, 'preferences_exclude'):
            for val in napari_config.preferences_exclude:
                schema['properties'].pop(val, None)
                values.pop(val, None)

        return schema, values

    def _restore_default_dialog(self):
        """Launches dialog to confirm restore settings choice."""
        prev = QApplication.instance().testAttribute(
            Qt.ApplicationAttribute.AA_DontUseNativeDialogs
        )
        QApplication.instance().setAttribute(
            Qt.ApplicationAttribute.AA_DontUseNativeDialogs, True
        )

        response = QMessageBox.question(
            self,
            trans._('Restore Settings'),
            trans._('Are you sure you want to restore default settings?'),
            QMessageBox.StandardButton.RestoreDefaults
            | QMessageBox.StandardButton.Cancel,
            QMessageBox.StandardButton.RestoreDefaults,
        )
        QApplication.instance().setAttribute(
            Qt.ApplicationAttribute.AA_DontUseNativeDialogs, prev
        )
        if response == QMessageBox.RestoreDefaults:
            self._settings.reset()

    def _restart_required_dialog(self):
        """Displays the dialog informing user a restart is required."""
        QMessageBox.information(
            self,
            trans._('Restart required'),
            trans._(
                'A restart is required for some new settings to have an effect.'
            ),
        )

    def closeEvent(self, event: 'QCloseEvent') -> None:
        event.accept()
        self.accept()

    def reject(self):
        """Restores the settings in place when dialog was launched."""
        self._settings.update(self._starting_values)

        # FIXME: this dialog should not need to know about the plugin manager
        if self._starting_pm_order:
            from napari.plugins import plugin_manager

            plugin_manager.set_call_order(self._starting_pm_order)
        super().reject()


def update_widget_state(name, widget):
    def _update_widget_state(event):
        value = event.value
        if isinstance(value, StrEnum):
            value = value.value
        widget.state = {name: value}

    return _update_widget_state



from typing import TYPE_CHECKING

from qtpy.QtCore import Qt
from qtpy.QtWidgets import (
    QCheckBox,
    QComboBox,
    QSlider,
)

from napari._qt.layer_controls.qt_layer_controls_base import QtLayerControls
from napari._qt.utils import qt_signals_blocked
from napari.layers.base._base_constants import Mode
from napari.utils.colormaps import AVAILABLE_COLORMAPS
from napari.utils.translations import trans

if TYPE_CHECKING:
    import napari.layers


class QtTracksControls(QtLayerControls):
    """Qt view and controls for the Tracks layer.

    Parameters
    ----------
    layer : napari.layers.Tracks
        An instance of a Tracks layer.

    Attributes
    ----------
    layer : layers.Tracks
        An instance of a Tracks layer.
    button_group : qtpy.QtWidgets.QButtonGroup
        Button group of points layer modes (ADD, PAN_ZOOM, SELECT).
    panzoom_button : napari._qt.widgets.qt_mode_button.QtModeRadioButton
        Button for pan/zoom mode.
    transform_button : napari._qt.widgets.qt_mode_button.QtModeRadioButton
        Button to select transform mode.

    """

    layer: 'napari.layers.Tracks'
    MODE = Mode
    PAN_ZOOM_ACTION_NAME = 'activate_tracks_pan_zoom_mode'
    TRANSFORM_ACTION_NAME = 'activate_tracks_transform_mode'

    def __init__(self, layer) -> None:
        super().__init__(layer)

        # NOTE(arl): there are no events fired for changing checkboxes
        self.layer.events.color_by.connect(self._on_color_by_change)
        self.layer.events.tail_width.connect(self._on_tail_width_change)
        self.layer.events.tail_length.connect(self._on_tail_length_change)
        self.layer.events.head_length.connect(self._on_head_length_change)
        self.layer.events.properties.connect(self._on_properties_change)
        self.layer.events.colormap.connect(self._on_colormap_change)

        # combo box for track coloring, we can get these from the properties
        # keys
        self.color_by_combobox = QComboBox()
        self.color_by_combobox.addItems(self.layer.properties_to_color_by)

        self.colormap_combobox = QComboBox()
        for name, colormap in AVAILABLE_COLORMAPS.items():
            display_name = colormap._display_name
            self.colormap_combobox.addItem(display_name, name)

        # slider for track head length
        self.head_length_slider = QSlider(Qt.Orientation.Horizontal)
        self.head_length_slider.setFocusPolicy(Qt.FocusPolicy.NoFocus)
        self.head_length_slider.setMinimum(0)
        self.head_length_slider.setMaximum(self.layer._max_length)
        self.head_length_slider.setSingleStep(1)

        # slider for track tail length
        self.tail_length_slider = QSlider(Qt.Orientation.Horizontal)
        self.tail_length_slider.setFocusPolicy(Qt.FocusPolicy.NoFocus)
        self.tail_length_slider.setMinimum(1)
        self.tail_length_slider.setMaximum(self.layer._max_length)
        self.tail_length_slider.setSingleStep(1)

        # slider for track edge width
        self.tail_width_slider = QSlider(Qt.Orientation.Horizontal)
        self.tail_width_slider.setFocusPolicy(Qt.FocusPolicy.NoFocus)
        self.tail_width_slider.setMinimum(1)
        self.tail_width_slider.setMaximum(int(2 * self.layer._max_width))
        self.tail_width_slider.setSingleStep(1)

        # checkboxes for display
        self.id_checkbox = QCheckBox()
        self.tail_checkbox = QCheckBox()
        self.tail_checkbox.setChecked(True)
        self.graph_checkbox = QCheckBox()
        self.graph_checkbox.setChecked(True)

        self.tail_width_slider.valueChanged.connect(self.change_tail_width)
        self.tail_length_slider.valueChanged.connect(self.change_tail_length)
        self.head_length_slider.valueChanged.connect(self.change_head_length)
        self.tail_checkbox.stateChanged.connect(self.change_display_tail)
        self.id_checkbox.stateChanged.connect(self.change_display_id)
        self.graph_checkbox.stateChanged.connect(self.change_display_graph)
        self.color_by_combobox.currentTextChanged.connect(self.change_color_by)
        self.colormap_combobox.currentTextChanged.connect(self.change_colormap)

        self.layout().addRow(self.button_grid)
        self.layout().addRow(self.opacityLabel, self.opacitySlider)
        self.layout().addRow(trans._('blending:'), self.blendComboBox)
        self.layout().addRow(trans._('color by:'), self.color_by_combobox)
        self.layout().addRow(trans._('colormap:'), self.colormap_combobox)
        self.layout().addRow(trans._('tail width:'), self.tail_width_slider)
        self.layout().addRow(trans._('tail length:'), self.tail_length_slider)
        self.layout().addRow(trans._('head length:'), self.head_length_slider)
        self.layout().addRow(trans._('tail:'), self.tail_checkbox)
        self.layout().addRow(trans._('show ID:'), self.id_checkbox)
        self.layout().addRow(trans._('graph:'), self.graph_checkbox)

        self._on_tail_length_change()
        self._on_tail_width_change()
        self._on_colormap_change()
        self._on_color_by_change()

    def _on_tail_width_change(self):
        """Receive layer model track line width change event and update slider."""
        with self.layer.events.tail_width.blocker():
            value = int(2 * self.layer.tail_width)
            self.tail_width_slider.setValue(value)

    def _on_tail_length_change(self):
        """Receive layer model track line width change event and update slider."""
        with self.layer.events.tail_length.blocker():
            value = self.layer.tail_length
            self.tail_length_slider.setValue(value)

    def _on_head_length_change(self):
        """Receive layer model track line width change event and update slider."""
        with self.layer.events.head_length.blocker():
            value = self.layer.head_length
            self.head_length_slider.setValue(value)

    def _on_properties_change(self):
        """Change the properties that can be used to color the tracks."""
        with qt_signals_blocked(self.color_by_combobox):
            self.color_by_combobox.clear()
            self.color_by_combobox.addItems(self.layer.properties_to_color_by)
        self._on_color_by_change()

    def _on_colormap_change(self):
        """Receive layer model colormap change event and update combobox."""
        with self.layer.events.colormap.blocker():
            self.colormap_combobox.setCurrentIndex(
                self.colormap_combobox.findData(self.layer.colormap)
            )

    def _on_color_by_change(self):
        """Receive layer model color_by change event and update combobox."""
        with self.layer.events.color_by.blocker():
            color_by = self.layer.color_by

            idx = self.color_by_combobox.findText(
                color_by, Qt.MatchFlag.MatchFixedString
            )
            self.color_by_combobox.setCurrentIndex(idx)

    def change_tail_width(self, value):
        """Change track line width of shapes on the layer model.

        Parameters
        ----------
        value : float
            Line width of track tails.
        """
        self.layer.tail_width = float(value) / 2.0

    def change_tail_length(self, value):
        """Change edge line backward length of shapes on the layer model.

        Parameters
        ----------
        value : int
            Line length of track tails.
        """
        self.layer.tail_length = value

    def change_head_length(self, value):
        """Change edge line forward length of shapes on the layer model.

        Parameters
        ----------
        value : int
            Line length of track tails.
        """
        self.layer.head_length = value

    def change_display_tail(self, state):
        self.layer.display_tail = self.tail_checkbox.isChecked()

    def change_display_id(self, state):
        self.layer.display_id = self.id_checkbox.isChecked()

    def change_display_graph(self, state):
        self.layer.display_graph = self.graph_checkbox.isChecked()

    def change_color_by(self, value: str):
        self.layer.color_by = value

    def change_colormap(self, colormap: str):
        self.layer.colormap = self.colormap_combobox.currentData()



import pytest

from napari._app_model import get_app_model
from napari._app_model.actions._layerlist_context_actions import (
    LAYERLIST_CONTEXT_ACTIONS,
)


@pytest.mark.parametrize('layer_action', LAYERLIST_CONTEXT_ACTIONS)
def test_layer_actions_ctx_menu_execute_command(
    layer_action, make_napari_viewer
):
    """
    Test layer context menu actions via app-model `execute_command`.

    Note:
        This test is here only to ensure app-model action dispatch mechanism
        is working for these actions (which use the `_provide_active_layer_list` provider).

        To check a set of functional tests related to these actions you can
        see: https://github.com/napari/napari/blob/main/napari/layers/_tests/test_layer_actions.py
    """
    app = get_app_model()
    make_napari_viewer()
    command_id = layer_action.id

    if command_id == 'napari.layer.merge_stack':
        with pytest.raises(IndexError, match=r'images list is empty'):
            app.commands.execute_command(command_id)
    elif command_id in [
        'napari.layer.link_selected_layers',
        'napari.layer.unlink_selected_layers',
    ]:
        with pytest.raises(ValueError, match=r'at least one'):
            app.commands.execute_command(command_id)
    else:
        app.commands.execute_command(command_id)



import numpy as np
from qtpy.QtCore import QModelIndex, QPoint, Qt
from qtpy.QtWidgets import QLineEdit, QStyleOptionViewItem

from napari._qt.containers import QtLayerList
from napari._qt.containers._layer_delegate import LayerDelegate
from napari.components import LayerList
from napari.layers import Image, Shapes


def test_set_layer_invisible_makes_item_unchecked(qtbot):
    view, image = make_qt_layer_list_with_layer(qtbot)
    assert image.visible
    assert check_state_at_layer_index(view, 0) == Qt.CheckState.Checked

    image.visible = False

    assert check_state_at_layer_index(view, 0) == Qt.CheckState.Unchecked


def test_set_item_unchecked_makes_layer_invisible(qtbot):
    view, image = make_qt_layer_list_with_layer(qtbot)
    assert check_state_at_layer_index(view, 0) == Qt.CheckState.Checked
    assert image.visible

    view.model().setData(
        layer_to_model_index(view, 0),
        Qt.CheckState.Unchecked,
        Qt.ItemDataRole.CheckStateRole,
    )

    assert not image.visible


def test_alt_click_to_show_single_layer(qtbot):
    (
        image1,
        image2,
        image3,
        layers,
        view,
        delegate,
    ) = make_qt_layer_list_with_delegate(qtbot)

    # hide the middle-layer, image2 and ensure it's unchecked
    image2.visible = False
    assert check_state_at_layer_index(view, 1) == Qt.CheckState.Unchecked

    # ensure the other layers are visible & checked
    assert image3.visible
    assert check_state_at_layer_index(view, 0) == Qt.CheckState.Checked
    assert image1.visible
    assert check_state_at_layer_index(view, 2) == Qt.CheckState.Checked

    # alt-click state should be None
    assert delegate._alt_click_layer() is None

    # mock an alt-click on bottom-most layer, image1
    index = layer_to_model_index(view, 2)
    delegate._show_on_alt_click_hide_others(view.model(), index)

    # alt-click state should be set to image1
    assert delegate._alt_click_layer() == image1

    # only image1 should be shown, while image3, image2 be hidden
    assert not image3.visible
    assert check_state_at_layer_index(view, 0) == Qt.CheckState.Unchecked
    assert not image2.visible
    assert check_state_at_layer_index(view, 1) == Qt.CheckState.Unchecked
    assert image1.visible
    assert check_state_at_layer_index(view, 2) == Qt.CheckState.Checked


def test_second_alt_click_to_show_different_layer(qtbot):
    (
        image1,
        image2,
        image3,
        layers,
        view,
        delegate,
    ) = make_qt_layer_list_with_delegate(qtbot)

    # mock an alt-click on bottom-most layer, image1
    index = layer_to_model_index(view, 2)
    delegate._show_on_alt_click_hide_others(view.model(), index)

    # alt-click state should be set to image1
    assert delegate._alt_click_layer() == image1

    # image2 should be hidden
    assert not image2.visible
    assert check_state_at_layer_index(view, 1) == Qt.CheckState.Unchecked

    # mock an alt-click on middle layer, image2
    index2 = layer_to_model_index(view, 1)
    delegate._show_on_alt_click_hide_others(view.model(), index2)

    # alt-click state should be set to image2
    assert delegate._alt_click_layer() == image2

    # only image2 should be shown, while image3, image1 be hidden
    assert not image3.visible
    assert check_state_at_layer_index(view, 0) == Qt.CheckState.Unchecked
    assert not image1.visible
    assert check_state_at_layer_index(view, 2) == Qt.CheckState.Unchecked
    assert image2.visible
    assert check_state_at_layer_index(view, 1) == Qt.CheckState.Checked


def test_second_alt_click_to_restore_layer_state(qtbot):
    (
        image1,
        image2,
        image3,
        layers,
        view,
        delegate,
    ) = make_qt_layer_list_with_delegate(qtbot)

    # mock an alt-click on bottom-most layer, image1
    index = layer_to_model_index(view, 2)
    delegate._show_on_alt_click_hide_others(view.model(), index)

    # add a layer (will be at position 0)
    image4 = Image(np.zeros((4, 3)))
    layers.append(image4)
    assert image4.visible

    # remove a layer (image3, which has been pushed down to position 1
    layers.pop(1)

    # mock second alt-click on image1, which should restore initial state
    delegate._show_on_alt_click_hide_others(view.model(), index)

    # image4 should remain visible (not part of initial state)
    # image2 should be not visible--that was the initial state
    assert image4.visible
    assert image1.visible
    assert not image2.visible

    # alt-click state should be cleared
    assert delegate._alt_click_layer() is None


def test_contextual_menu_updates_selection_ctx_keys(monkeypatch, qtbot):
    shapes_layer = Shapes()
    layer_list = LayerList()
    layer_list._create_contexts()
    layer_list.append(shapes_layer)
    view = QtLayerList(layer_list)
    qtbot.addWidget(view)
    delegate = view.itemDelegate()
    assert not layer_list[0].data

    layer_list.selection.add(shapes_layer)
    index = layer_to_model_index(view, 0)
    assert layer_list._selection_ctx_keys.num_selected_shapes_layers == 1
    assert layer_list._selection_ctx_keys.selected_empty_shapes_layer

    monkeypatch.setattr(
        'app_model.backends.qt.QModelMenu.exec_', lambda self, x: x
    )

    delegate.show_context_menu(
        index, view.model(), QPoint(10, 10), parent=view
    )
    assert not delegate._context_menu.findAction(
        'napari.layer.convert_to_labels'
    ).isEnabled()

    layer_list[0].add(np.array(([0, 0], [0, 10], [10, 10], [10, 0])))
    assert layer_list[0].data
    delegate.show_context_menu(
        index, view.model(), QPoint(10, 10), parent=view
    )
    assert delegate._context_menu.findAction(
        'napari.layer.convert_to_labels'
    ).isEnabled()


def make_qt_layer_list_with_delegate(qtbot):
    image1 = Image(np.zeros((4, 3)))
    image2 = Image(np.zeros((4, 3)))
    image3 = Image(np.zeros((4, 3)))

    layers = LayerList([image1, image2, image3])
    # this will make the list have image2 on top of image1
    view = QtLayerList(layers)
    qtbot.addWidget(view)

    delegate = LayerDelegate()
    return image1, image2, image3, layers, view, delegate


def make_qt_layer_list_with_layer(qtbot) -> tuple[QtLayerList, Image]:
    image = Image(np.zeros((4, 3)))
    layers = LayerList([image])
    view = QtLayerList(layers)
    qtbot.addWidget(view)
    return view, image


def layer_to_model_index(view: QtLayerList, layer_index: int) -> QModelIndex:
    return view.model().index(layer_index, 0, view.rootIndex())


def check_state_at_layer_index(
    view: QtLayerList, layer_index: int
) -> Qt.CheckState:
    model_index = layer_to_model_index(view, layer_index)
    value = view.model().data(model_index, Qt.ItemDataRole.CheckStateRole)
    # The data method returns integer value of the enum in some cases, so
    # ensure it has the enum type for more explicit assertions.
    return Qt.CheckState(value)


def test_createEditor(qtbot):
    view, image = make_qt_layer_list_with_layer(qtbot)
    model_index = layer_to_model_index(view, 0)
    delegate = view.itemDelegate()
    editor = delegate.createEditor(view, QStyleOptionViewItem(), model_index)
    assert isinstance(editor, QLineEdit)
    delegate.setEditorData(editor, model_index)
    assert editor.text() == image.name



from qtpy.QtCore import QModelIndex, Qt

from napari._qt.containers import (
    AxisList,
    AxisModel,
    QtAxisListModel,
    QtListView,
)
from napari.components import Dims

FLAGS = (
    Qt.ItemFlag.ItemIsSelectable
    | Qt.ItemFlag.ItemIsEditable
    | Qt.ItemFlag.ItemIsUserCheckable
    | Qt.ItemFlag.ItemIsEnabled
    | Qt.ItemFlag.ItemNeverHasChildren
    | Qt.ItemFlag.ItemIsDragEnabled
)


def test_axismodel():
    dims = Dims()
    axismodel = AxisModel(dims, 0)
    assert axismodel == 0
    assert axismodel.rollable

    axismodel.rollable = False
    assert not axismodel.rollable
    assert not dims.rollable[0]


def test_AxisList():
    # from list
    dims = Dims()
    axes = [AxisModel(dims, axis) for axis in dims.order]
    axislist = AxisList(axes)
    assert all(axis == idx for idx, axis in enumerate(axislist))

    # from_dims
    axislist = AxisList.from_dims(dims)
    assert len(axislist) == 2
    for idx, axis in enumerate(axislist):
        assert axis == idx


def test_QtAxisListModel_data(qtbot):
    dims, axislist, listview, axislistmodel = make_QtAxisListModel(qtbot)
    assert all(
        axislistmodel.data(
            axislistmodel.index(idx), role=Qt.ItemDataRole.DisplayRole
        )
        == axislist[idx]
        for idx in dims.order
    )
    assert all(
        axislistmodel.data(
            axislistmodel.index(idx),
            role=Qt.ItemDataRole.TextAlignmentRole,
        )
        == Qt.AlignCenter
        for idx in dims.order
    )
    assert all(
        (
            axislistmodel.data(
                axislistmodel.index(idx),
                role=Qt.ItemDataRole.CheckStateRole,
            ),
            axislist[idx].rollable,
        )
        for idx in dims.order
    )

    # setData
    idx = 1
    with qtbot.waitSignal(axislistmodel.dataChanged, timeout=100):
        assert axislistmodel.setData(
            axislistmodel.index(idx),
            Qt.CheckState.Checked,
            role=Qt.ItemDataRole.CheckStateRole,
        )
    assert dims.rollable[idx]

    new_name = 'new_name'
    with qtbot.waitSignal(axislistmodel.dataChanged, timeout=100):
        assert axislistmodel.setData(
            axislistmodel.index(idx), new_name, role=Qt.ItemDataRole.EditRole
        )
    assert dims.axis_labels[idx] == new_name


def test_QtAxisListModel_flags(qtbot):
    dims, axislist, listview, axislistmodel = make_QtAxisListModel(qtbot)
    assert axislistmodel.flags(QModelIndex()) == Qt.ItemFlag.ItemIsDropEnabled
    flags = [
        axislistmodel.flags(axislistmodel.index(idx)) for idx in dims.order
    ]
    ref_flags = [FLAGS for idx in dims.order]
    assert flags == ref_flags


def make_QtAxisListModel(qtbot) -> tuple[Dims, AxisList, QtAxisListModel]:
    dims = Dims()
    dims.rollable = [True, False]
    axislist = AxisList.from_dims(dims)
    view = QtListView(axislist)
    axislistmodel = view.model()
    return dims, axislist, view, axislistmodel



import inspect
import os
import site
from textwrap import indent
from typing import TYPE_CHECKING, ClassVar

from napari._pydantic_compat import BaseSettings, Field, PrivateAttr
from napari.utils.misc import ROOT_DIR
from napari.utils.translations import trans

try:
    from rich import print
except ModuleNotFoundError:
    print(
        trans._(
            'TIP: run `pip install rich` for much nicer event debug printout.'
        )
    )
try:
    import dotenv
except ModuleNotFoundError:
    dotenv = None  # type: ignore

if TYPE_CHECKING:
    from napari.utils.events.event import Event


class EventDebugSettings(BaseSettings):
    """Parameters controlling how event debugging logs appear.

    To enable Event debugging:
        1. pip install rich pydantic[dotenv]
        2. export NAPARI_DEBUG_EVENTS=1  # or modify the .env_sample file
        3. see .env_sample file for ways to set these fields here.
    """

    # event emitters (e.g. 'Shapes') and event names (e.g. 'set_data')
    # to include/exclude when printing events.
    include_emitters: set[str] = Field(default_factory=set)
    include_events: set[str] = Field(default_factory=set)
    exclude_emitters: set[str] = Field(
        default_factory=lambda: {'TransformChain', 'Context'}
    )
    exclude_events: set[str] = Field(
        default_factory=lambda: {'status', 'position'}
    )
    # stack depth to show
    stack_depth: int = 20
    # how many sub-emit nesting levels to show
    # (i.e. events that get triggered by other events)
    nesting_allowance: int = 0

    _cur_depth: ClassVar[int] = PrivateAttr(0)

    class Config:
        env_prefix = 'event_debug_'
        env_file = '.env' if dotenv is not None else ''


_SETTINGS = EventDebugSettings()
_SP = site.getsitepackages()[0]
_STD_LIB = site.__file__.rsplit(os.path.sep, 1)[0]


def _shorten_fname(fname: str) -> str:
    """Reduce extraneous stuff from filenames"""
    fname = fname.replace(_SP, '.../site-packages')
    fname = fname.replace(_STD_LIB, '.../python')
    return fname.replace(ROOT_DIR, 'napari')


def log_event_stack(event: 'Event', cfg: EventDebugSettings = _SETTINGS):
    """Print info about what caused this event to be emitted.s"""

    if cfg.include_events:
        if event.type not in cfg.include_events:
            return
    elif event.type in cfg.exclude_events:
        return

    source = type(event.source).__name__
    if cfg.include_emitters:
        if source not in cfg.include_emitters:
            return
    elif source in cfg.exclude_emitters:
        return

    # get values being emitted
    vals = ','.join(f'{k}={v}' for k, v in event._kwargs.items())
    # show event type and source
    lines = [f'{source}.events.{event.type}({vals})']
    # climb stack and show what caused it.
    # note, we start 2 frames back in the stack, one frame for *this* function
    # and the second frame for the EventEmitter.__call__ function (where this
    # function was likely called).
    call_stack = inspect.stack(0)
    for frame in call_stack[2 : 2 + cfg.stack_depth]:
        fname = _shorten_fname(frame.filename)
        obj = ''
        if 'self' in frame.frame.f_locals:
            obj = type(frame.frame.f_locals['self']).__name__ + '.'
        ln = f'  "{fname}", line {frame.lineno}, in {obj}{frame.function}'
        lines.append(ln)
    lines.append('')

    # find the first caller in the call stack
    for f in reversed(call_stack):
        if 'self' in f.frame.f_locals:
            obj_type = type(f.frame.f_locals['self'])
            module = obj_type.__module__ or ''
            if module.startswith('napari'):
                trigger = f'{obj_type.__name__}.{f.function}()'
                lines.insert(1, f'  was triggered by {trigger}, via:')
                break

    # separate groups of events
    if not cfg._cur_depth:
        lines = ['─' * 79, '', *lines]
    elif not cfg.nesting_allowance:
        return

    # log it
    print(indent('\n'.join(lines), '  ' * cfg._cur_depth))

    # spy on nested events...
    # (i.e. events that were emitted while another was being emitted)
    def _pop_source():
        cfg._cur_depth -= 1
        return event._sources.pop()

    event._pop_source = _pop_source
    cfg._cur_depth += 1



"""Monitor service."""

from napari.components.experimental.monitor._monitor import monitor
from napari.components.experimental.monitor._utils import numpy_dumps

__all__ = ['monitor', 'numpy_dumps']



from unittest.mock import Mock

import pytest

from napari._app_model.context._context import (
    ContextMapping,
    create_context,
    get_context,
)


def test_simple_mapping():
    data = {'a': 1, 'b': 2}
    mapping = ContextMapping(data)
    assert mapping['a'] == 1
    assert mapping['b'] == 2
    assert 'a' in mapping
    assert list(mapping) == ['a', 'b']
    assert len(mapping) == 2


def test_missed_key():
    data = {'a': 1, 'b': 2}
    mapping = ContextMapping(data)
    with pytest.raises(KeyError):
        mapping['c']


def test_callable_value():
    data = {'a': 1, 'b': Mock(return_value=2)}
    mapping = ContextMapping(data)
    assert mapping['a'] == 1
    assert mapping['b'] == 2
    assert mapping['b'] == 2  # it is important to use [] twice in this test
    data['b'].assert_called_once()


def test_context_integration():
    obj = {1, 2, 3}
    ctx = create_context(obj)
    ctx['a'] = 1
    ctx['b'] = Mock(return_value=2)

    assert isinstance(get_context(obj), ContextMapping)
    mapping = get_context(obj)
    assert mapping['a'] == 1
    assert mapping['b'] == 2
    assert mapping['b'] == 2  # it is important to use [] twice in this test
    ctx['b'].assert_called_once()

    mapping2 = get_context(obj)
    assert mapping2['b'] == 2
    assert ctx['b'].call_count == 2
    assert mapping2['b'] == 2
    assert ctx['b'].call_count == 2



from unittest.mock import Mock

import numpy as np

from napari.layers.base._base_mouse_bindings import _rotate_with_box
from napari.utils.transforms import Affine


def test_interaction_box_rotation():
    layer = Mock(affine=Affine())
    initial_affine = Affine()
    initial_mouse_pos = Mock()
    # rotation handle is 8th
    initial_handle_coords = np.asarray(
        [
            [0, 0],
            [1, 1],
            [2, 2],
            [3, 3],
            [4, 4],
            [5, 5],
            [6, 6],
            [7, 7],
            [6, 3],
        ],
        dtype=np.float32,
    )
    initial_center = np.asarray([3, 3], dtype=np.float32)
    mouse_pos = np.asarray([6, 5], dtype=np.float32)
    event = Mock(dims_displayed=[0, 1], modifiers=[None])
    _rotate_with_box(
        layer,
        initial_affine,
        initial_mouse_pos,
        initial_handle_coords,
        initial_center,
        mouse_pos,
        event,
    )
    # should be approximately 33 degrees
    assert np.allclose(layer.affine.rotate, Affine(rotate=33.69).rotate)


def test_interaction_box_fixed_rotation():
    layer = Mock(affine=Affine())
    initial_affine = Affine()
    initial_mouse_pos = Mock()
    # rotation handle is 8th
    initial_handle_coords = np.asarray(
        [
            [0, 0],
            [1, 1],
            [2, 2],
            [3, 3],
            [4, 4],
            [5, 5],
            [6, 6],
            [7, 7],
            [6, 3],
        ],
        dtype=np.float32,
    )
    initial_center = np.asarray([3, 3], dtype=np.float32)
    mouse_pos = np.asarray([6, 5], dtype=np.float32)
    # use Shift to snap rotation to steps of 45 degrees
    event = Mock(dims_displayed=[0, 1], modifiers=['Shift'])
    _rotate_with_box(
        layer,
        initial_affine,
        initial_mouse_pos,
        initial_handle_coords,
        initial_center,
        mouse_pos,
        event,
    )
    # should be 45 degrees
    assert np.allclose(layer.affine.rotate, Affine(rotate=45).rotate)



import weakref
from functools import partial

import pytest

from napari.utils.events import EventEmitter


def test_event_blocker_count_none():
    """Test event emitter block counter with no emission."""
    e = EventEmitter(type_name='test')
    with e.blocker() as block:
        pass
    assert block.count == 0


def test_event_blocker_count():
    """Test event emitter block counter with emission."""
    e = EventEmitter(type_name='test')
    with e.blocker() as block:
        e()
        e()
        e()
    assert block.count == 3


def test_weakref_event_emitter():
    """
    We are testing that an event blocker does not keep hard reference to
    the object we are blocking, especially if it's a bound method.

    The reason it used to keep references is to get the count of how many time
    a callback was blocked, but if the object does not exists, then the bound method
    does not and thus there is no way to ask for it's count.

    so we can keep only weak refs.

    """
    e = EventEmitter(type_name='test_weak')

    class Obj:
        def cb(self):
            pass

    o = Obj()
    ref_o = weakref.ref(o)

    e.connect(o.cb)

    #
    with e.blocker(o.cb):
        e()

    del o
    assert ref_o() is None


@pytest.mark.parametrize('disconnect_and_should_be_none', [True, False])
def test_weakref_event_emitter_cb(disconnect_and_should_be_none):
    """

    Note that as above but with pure callback, We keep a reference to it, the
    reason is that unlike with bound method, the callback may be a closure and
    may not stick around.

    We thus expect the wekref to be None only if explicitely disconnected

    """
    e = EventEmitter(type_name='test_weak')

    def cb(self):
        pass

    ref_cb = weakref.ref(cb)

    e.connect(cb)

    with e.blocker(cb):
        e()

    if disconnect_and_should_be_none:
        e.disconnect(cb)
        del cb
        assert ref_cb() is None
    else:
        del cb
        assert ref_cb() is not None


def test_error_on_connect():
    """Check that connections happen correctly even on decorated methods.

    Some decorators will alter method.__name__, so that obj.method
    will not be equal to getattr(obj, obj.method.__name__). We check here
    that event binding will be correct even in these situations.
    """

    def rename(newname):
        def decorator(f):
            f.__name__ = newname
            return f

        return decorator

    class Test:
        def __init__(self) -> None:
            self.m1, self.m2, self.m4 = 0, 0, 0

        @rename('nonexist')
        def meth1(self, _event):
            self.m1 += 1

        @rename('meth1')
        def meth2(self, _event):
            self.m2 += 1

        def meth3(self):
            pass

        def meth4(self, _event):
            self.m4 += 1

    t = Test()

    e = EventEmitter(type_name='test')

    e.connect(t.meth1)
    e()
    assert (t.m1, t.m2) == (1, 0)

    e.connect(t.meth2)
    e()
    assert (t.m1, t.m2) == (2, 1)

    meth = t.meth3

    t.meth3 = 'aaaa'

    with pytest.raises(RuntimeError):
        e.connect(meth)

    e.connect(t.meth4)
    assert t.m4 == 0
    e()
    assert t.m4 == 1
    t.meth4 = None
    with pytest.warns(RuntimeWarning, match='Problem with function'):
        e()
    assert t.m4 == 1


def test_event_order_func():
    res_li = []

    def fun1():
        res_li.append(1)

    def fun2(val):
        res_li.append(val)

    def fun3():
        res_li.append(3)

    def fun4():
        res_li.append(4)

    def fun5(val):
        res_li.append(val)

    def fun6(val):
        res_li.append(val)

    fun1.__module__ = 'napari.test.sample'
    fun3.__module__ = 'napari.test.sample'
    fun5.__module__ = 'napari.test.sample'

    e = EventEmitter(type_name='test')
    e.connect(fun1)
    e.connect(partial(fun2, val=2))
    e()
    assert res_li == [1, 2]
    res_li = []
    e.connect(fun3)
    e()
    assert res_li == [1, 3, 2]
    res_li = []
    e.connect(fun4)
    e()
    assert res_li == [1, 3, 2, 4]
    res_li = []
    e.connect(partial(fun5, val=5), position='first')
    e()
    assert res_li == [5, 1, 3, 2, 4]
    res_li = []
    e.connect(partial(fun6, val=6), position='first')
    e()
    assert res_li == [5, 1, 3, 6, 2, 4]


def test_event_order_methods():
    res_li = []

    class Test:
        def fun1(self):
            res_li.append(1)

        def fun2(self):
            res_li.append(2)

    class Test2:
        def fun3(self):
            res_li.append(3)

        def fun4(self):
            res_li.append(4)

    Test.__module__ = 'napari.test.sample'

    t1 = Test()
    t2 = Test2()

    e = EventEmitter(type_name='test')
    e.connect(t1.fun1)
    e.connect(t2.fun3)
    e()
    assert res_li == [1, 3]
    res_li = []
    e.connect(t1.fun2)
    e.connect(t2.fun4)
    e()
    assert res_li == [1, 2, 3, 4]


def test_no_event_arg():
    class TestOb:
        def __init__(self) -> None:
            self.count = 0

        def fun(self):
            self.count += 1

    count = [0]

    def simple_fun():
        count[0] += 1

    t = TestOb()

    e = EventEmitter(type_name='test')
    e.connect(t.fun)
    e.connect(simple_fun)
    e()
    assert t.count == 1
    assert count[0] == 1


def test_to_many_positional():
    class TestOb:
        def fun(self, a, b, c=1):
            pass

    def simple_fun(a, b):
        pass

    t = TestOb()

    e = EventEmitter(type_name='test')
    with pytest.raises(RuntimeError):
        e.connect(t.fun)
    with pytest.raises(RuntimeError):
        e.connect(simple_fun)


def test_disconnect_object():
    count_list = []

    def fun1():
        count_list.append(1)

    class TestOb:
        call_list_1 = []
        call_list_2 = []

        def fun1(self):
            self.call_list_1.append(1)

        def fun2(self):
            self.call_list_2.append(1)

    t = TestOb()

    e = EventEmitter(type_name='test')
    e.connect(t.fun1)
    e.connect(t.fun2)
    e.connect(fun1)
    e()

    assert t.call_list_1 == [1]
    assert t.call_list_2 == [1]
    assert count_list == [1]

    e.disconnect(t)
    e()

    assert t.call_list_1 == [1]
    assert t.call_list_2 == [1]
    assert count_list == [1, 1]


def test_weakref_disconnect():
    class TestOb:
        call_list_1 = []

        def fun1(self):
            self.call_list_1.append(1)

        def fun2(self, event):
            self.call_list_1.append(2)

    t = TestOb()

    e = EventEmitter(type_name='test')
    e.connect(t.fun1)
    e()

    assert t.call_list_1 == [1]
    e.disconnect((weakref.ref(t), 'fun1'))
    e()
    assert t.call_list_1 == [1]
    e.connect(t.fun2)
    e()
    assert t.call_list_1 == [1, 2]


def test_none_disconnect():
    count_list = []

    def fun1():
        count_list.append(1)

    def fun2(event):
        count_list.append(2)

    e = EventEmitter(type_name='test')
    e.connect(fun1)
    e()
    assert count_list == [1]
    e.disconnect(None)
    e()
    assert count_list == [1]
    e.connect(fun2)
    e()
    assert count_list == [1, 2]



import numpy as np
import pytest

from napari._qt.widgets.qt_highlight_preview import (
    QtHighlightPreviewWidget,
    QtStar,
    QtTriangle,
)


@pytest.fixture
def star_widget(qtbot):
    def _star_widget(**kwargs):
        widget = QtStar(**kwargs)
        widget.show()
        qtbot.addWidget(widget)

        return widget

    return _star_widget


@pytest.fixture
def triangle_widget(qtbot):
    def _triangle_widget(**kwargs):
        widget = QtTriangle(**kwargs)
        widget.show()
        qtbot.addWidget(widget)

        return widget

    return _triangle_widget


@pytest.fixture
def highlight_preview_widget(qtbot):
    def _highlight_preview_widget(**kwargs):
        widget = QtHighlightPreviewWidget(**kwargs)
        widget.show()
        qtbot.addWidget(widget)

        return widget

    return _highlight_preview_widget


# QtStar
# ----------------------------------------------------------------------------


def test_qt_star_defaults(star_widget):
    star_widget()


def test_qt_star_value(star_widget):
    widget = star_widget(value=5)
    assert widget.value() <= 5

    widget = star_widget()
    widget.setValue(5)
    assert widget.value() == 5


# QtTriangle
# ----------------------------------------------------------------------------


def test_qt_triangle_defaults(triangle_widget):
    triangle_widget()


def test_qt_triangle_value(triangle_widget):
    widget = triangle_widget(value=5)
    assert widget.value() <= 5

    widget = triangle_widget()
    widget.setValue(5)
    assert widget.value() == 5


def test_qt_triangle_minimum(triangle_widget):
    minimum = 1
    widget = triangle_widget(min_value=minimum)
    assert widget.minimum() == minimum
    assert widget.value() >= minimum

    widget = triangle_widget()
    widget.setMinimum(2)
    assert widget.minimum() == 2
    assert widget.value() == 2


def test_qt_triangle_maximum(triangle_widget):
    maximum = 10
    widget = triangle_widget(max_value=maximum)
    assert widget.maximum() == maximum
    assert widget.value() <= maximum

    widget = triangle_widget()
    widget.setMaximum(20)
    assert widget.maximum() == 20


def test_qt_triangle_signal(qtbot, triangle_widget):
    widget = triangle_widget()

    with qtbot.waitSignal(widget.valueChanged, timeout=500):
        widget.setValue(7)

    with qtbot.waitSignal(widget.valueChanged, timeout=500):
        widget.setValue(-5)


# QtHighlightPreviewWidget
# ----------------------------------------------------------------------------


def test_qt_highlight_preview_widget_defaults(
    highlight_preview_widget,
):
    highlight_preview_widget()


def test_qt_highlight_preview_widget_description(
    highlight_preview_widget,
):
    description = 'Some text'
    widget = highlight_preview_widget(description=description)
    assert widget.description() == description

    widget = highlight_preview_widget()
    widget.setDescription(description)
    assert widget.description() == description


def test_qt_highlight_preview_widget_unit(highlight_preview_widget):
    unit = 'CM'
    widget = highlight_preview_widget(unit=unit)
    assert widget.unit() == unit

    widget = highlight_preview_widget()
    widget.setUnit(unit)
    assert widget.unit() == unit


def test_qt_highlight_preview_widget_minimum(
    highlight_preview_widget,
):
    minimum = 5
    widget = highlight_preview_widget(min_value=minimum)
    assert widget.minimum() == minimum
    assert widget._thickness_value >= minimum
    assert widget.value()['highlight_thickness'] >= minimum

    widget = highlight_preview_widget()
    widget.setMinimum(3)
    assert widget.minimum() == 3
    assert widget._thickness_value == 3
    assert widget.value()['highlight_thickness'] == 3
    assert widget._slider.minimum() == 3
    assert widget._slider_min_label.text() == '3'
    assert widget._triangle.minimum() == 3
    assert widget._lineedit.text() == '3'


def test_qt_highlight_preview_widget_minimum_invalid(
    highlight_preview_widget,
):
    widget = highlight_preview_widget()

    with pytest.raises(ValueError, match='must be smaller than'):
        widget.setMinimum(60)


def test_qt_highlight_preview_widget_maximum(
    highlight_preview_widget,
):
    maximum = 10
    widget = highlight_preview_widget(max_value=maximum)

    assert widget.maximum() == maximum
    assert widget._thickness_value <= maximum
    assert widget.value()['highlight_thickness'] <= maximum

    widget = highlight_preview_widget(
        value={
            'highlight_thickness': 6,
            'highlight_color': [0.0, 0.6, 1.0, 1.0],
        }
    )
    widget.setMaximum(20)
    assert widget.maximum() == 20
    assert widget._slider.maximum() == 20
    assert widget._triangle.maximum() == 20
    assert widget._slider_max_label.text() == '20'

    assert widget._thickness_value == 6
    assert widget.value()['highlight_thickness'] == 6
    widget.setMaximum(5)
    assert widget.maximum() == 5
    assert widget._thickness_value == 5
    assert widget.value()['highlight_thickness'] == 5
    assert widget._slider.maximum() == 5
    assert widget._triangle.maximum() == 5
    assert widget._lineedit.text() == '5'
    assert widget._slider_max_label.text() == '5'


def test_qt_highlight_preview_widget_maximum_invalid(
    highlight_preview_widget,
):
    widget = highlight_preview_widget()

    with pytest.raises(ValueError, match='must be larger than'):
        widget.setMaximum(-5)


def test_qt_highlight_preview_widget_value(highlight_preview_widget):
    widget = highlight_preview_widget(
        value={
            'highlight_thickness': 5,
            'highlight_color': [0.0, 0.6, 1.0, 1.0],
        }
    )
    assert widget._thickness_value <= 5
    assert widget.value()['highlight_thickness'] <= 5
    assert widget._color_value == [0.0, 0.6, 1.0, 1.0]
    assert widget.value()['highlight_color'] == [0.0, 0.6, 1.0, 1.0]

    widget = highlight_preview_widget()
    widget.setValue(
        {'highlight_thickness': 5, 'highlight_color': [0.6, 0.6, 1.0, 1.0]}
    )
    assert widget._thickness_value == 5
    assert widget.value()['highlight_thickness'] == 5
    assert np.array_equal(
        np.array(widget._color_value, dtype=np.float32),
        np.array([0.6, 0.6, 1.0, 1.0], dtype=np.float32),
    )
    assert np.array_equal(
        np.array(widget.value()['highlight_color'], dtype=np.float32),
        np.array([0.6, 0.6, 1.0, 1.0], dtype=np.float32),
    )


def test_qt_highlight_preview_widget_value_invalid(
    qtbot, highlight_preview_widget
):
    widget = highlight_preview_widget()
    widget.setMaximum(50)
    widget.setValue(
        {'highlight_thickness': 51, 'highlight_color': [0.0, 0.6, 1.0, 1.0]}
    )
    assert widget.value()['highlight_thickness'] == 50
    assert widget._lineedit.text() == '50'

    widget.setMinimum(5)
    widget.setValue(
        {'highlight_thickness': 1, 'highlight_color': [0.0, 0.6, 1.0, 1.0]}
    )
    assert widget.value()['highlight_thickness'] == 5
    assert widget._lineedit.text() == '5'


def test_qt_highlight_preview_widget_signal(qtbot, highlight_preview_widget):
    widget = highlight_preview_widget()

    with qtbot.waitSignal(widget.valueChanged, timeout=500):
        widget.setValue(
            {'highlight_thickness': 7, 'highlight_color': [0.0, 0.6, 1.0, 1.0]}
        )

    with qtbot.waitSignal(widget.valueChanged, timeout=500):
        widget.setValue(
            {
                'highlight_thickness': -5,
                'highlight_color': [0.0, 0.6, 1.0, 1.0],
            }
        )



from napari.layers.points import _points_key_bindings
from napari.layers.points.points import Points

# Note that importing _points_key_bindings is needed as the Points layer gets
# decorated with keybindings during that process, but it is not directly needed
# by our users and so is deleted below
del _points_key_bindings


__all__ = ['Points']



from collections.abc import Generator
from typing import (
    TYPE_CHECKING,
    Any,
    Callable,
    Optional,
    Union,
)

import numpy as np

from napari._pydantic_compat import errors, types

if TYPE_CHECKING:
    from decimal import Decimal

    from napari._pydantic_compat import ModelField

    Number = Union[int, float, Decimal]

# In numpy 2, the semantics of the copy argument in np.array changed
# so that copy=False errors if a copy is needed:
# https://numpy.org/devdocs/numpy_2_0_migration_guide.html#adapting-to-changes-in-the-copy-keyword
#
# In numpy 1, copy=False meant that a copy was avoided unless necessary,
# but would not error.
#
# In most usage like this use np.asarray instead, but sometimes we need
# to use some of the unique arguments of np.array (e.g. ndmin).
#
# This solution assumes numpy 1 by default, and switches to the numpy 2
# value for any release of numpy 2 on PyPI (including betas and RCs).
copy_if_needed: Optional[bool] = False
if np.lib.NumpyVersion(np.__version__) >= '2.0.0b1':
    copy_if_needed = None


class Array(np.ndarray):
    def __class_getitem__(cls, t):
        return type('Array', (Array,), {'__dtype__': t})

    @classmethod
    def __get_validators__(cls):
        yield cls.validate_type

    @classmethod
    def validate_type(cls, val):
        dtype = getattr(cls, '__dtype__', None)
        if isinstance(dtype, tuple):
            dtype, shape = dtype
        else:
            shape = ()

        result = np.array(
            val, dtype=dtype, copy=copy_if_needed, ndmin=len(shape)
        )

        if any(
            (shape[i] != -1 and shape[i] != result.shape[i])
            for i in range(len(shape))
        ):
            result = result.reshape(shape)
        return result


class NumberNotEqError(errors.PydanticValueError):
    code = 'number.not_eq'
    msg_template = 'ensure this value is not equal to {prohibited}'

    def __init__(self, *, prohibited: 'Number') -> None:
        super().__init__(prohibited=prohibited)


class ConstrainedInt(types.ConstrainedInt):
    """ConstrainedInt extension that adds not-equal"""

    ne: Optional[Union[int, list[int]]] = None

    @classmethod
    def __modify_schema__(cls, field_schema: dict[str, Any]) -> None:
        super().__modify_schema__(field_schema)
        if cls.ne is not None:
            f = 'const' if isinstance(cls.ne, int) else 'enum'
            field_schema['not'] = {f: cls.ne}

    @classmethod
    def __get_validators__(cls) -> Generator[Callable[..., Any], None, None]:
        yield from super().__get_validators__()
        yield cls.validate_ne

    @staticmethod
    def validate_ne(v: 'Number', field: 'ModelField') -> 'Number':
        field_type: ConstrainedInt = field.type_
        _ne = field_type.ne
        if _ne is not None and v in (_ne if isinstance(_ne, list) else [_ne]):
            raise NumberNotEqError(prohibited=field_type.ne)
        return v


def conint(
    *,
    strict: bool = False,
    gt: Optional[int] = None,
    ge: Optional[int] = None,
    lt: Optional[int] = None,
    le: Optional[int] = None,
    multiple_of: Optional[int] = None,
    ne: Optional[int] = None,
) -> type[int]:
    """Extended version of `pydantic.types.conint` that includes not-equal."""
    # use kwargs then define conf in a dict to aid with IDE type hinting
    namespace = {
        'strict': strict,
        'gt': gt,
        'ge': ge,
        'lt': lt,
        'le': le,
        'multiple_of': multiple_of,
        'ne': ne,
    }
    return type('ConstrainedIntValue', (ConstrainedInt,), namespace)



import numpy as np
import pint
import pytest
import zarr

from napari.components.layerlist import LayerList
from napari.layers import Image, Labels, Points, Shapes
from napari.layers._layer_actions import (
    _convert,
    _convert_dtype,
    _duplicate_layer,
    _hide_selected,
    _hide_unselected,
    _link_selected_layers,
    _merge_stack,
    _project,
    _show_selected,
    _show_unselected,
    _split_rgb,
    _split_stack,
    _toggle_visibility,
)

REG = pint.get_application_registry()


def test_split_stack():
    layer_list = LayerList()
    layer_list.append(Image(np.random.rand(8, 8, 8)))
    assert len(layer_list) == 1

    layer_list.selection.active = layer_list[0]
    _split_stack(layer_list)
    assert len(layer_list) == 8

    for idx in range(8):
        assert layer_list[idx].data.shape == (8, 8)


def test_split_rgb():
    layer_list = LayerList()
    layer_list.append(Image(np.random.random((48, 48, 3))))
    assert len(layer_list) == 1
    assert layer_list[0].rgb is True

    layer_list.selection.active = layer_list[0]
    _split_rgb(layer_list)
    assert len(layer_list) == 3

    for idx in range(3):
        assert layer_list[idx].data.shape == (48, 48)


def test_merge_stack():
    layer_list = LayerList()
    layer_list.append(Image(np.random.rand(8, 8)))
    layer_list.append(Image(np.random.rand(8, 8)))
    assert len(layer_list) == 2

    layer_list.selection.active = layer_list[0]
    layer_list.selection.add(layer_list[1])
    _merge_stack(layer_list)
    assert len(layer_list) == 1
    assert layer_list[0].data.shape == (2, 8, 8)


def test_toggle_visibility():
    """Test toggling visibility of a layer."""
    layer_list = LayerList()
    layer_list.append(Points([[0, 0]]))
    layer_list[0].visible = False

    layer_list.selection.active = layer_list[0]
    _toggle_visibility(layer_list)

    assert layer_list[0].visible is True


def test_toggle_visibility_with_linked_layers():
    """Test toggling visibility of a layer."""
    layer_list = LayerList()
    layer_list.append(Points([[0, 0]]))
    layer_list.append(Points([[0, 0]]))
    layer_list.append(Points([[0, 0]]))
    layer_list.append(Points([[0, 0]]))

    layer_list.selection.active = layer_list[0]
    layer_list.selection.add(layer_list[1])
    layer_list.selection.add(layer_list[2])

    _link_selected_layers(layer_list)

    layer_list[3].visible = False

    layer_list.selection.remove(layer_list[0])
    layer_list.selection.add(layer_list[3])

    _toggle_visibility(layer_list)

    assert layer_list[0].visible is False
    assert layer_list[1].visible is False
    assert layer_list[2].visible is False
    assert layer_list[3].visible is True


@pytest.mark.parametrize('layer_type', [Points, Shapes])
def test_duplicate_layers(layer_type):
    def _dummy():
        pass

    layer_list = LayerList()
    layer_list.append(layer_type([], name='test'))
    layer_list.selection.active = layer_list[0]
    layer_list[0].events.data.connect(_dummy)
    assert len(layer_list[0].events.data.callbacks) == 2
    assert len(layer_list) == 1
    _duplicate_layer(layer_list)
    assert len(layer_list) == 2
    assert layer_list[0].name == 'test'
    assert layer_list[1].name == 'test copy'
    assert layer_list[1].events.source is layer_list[1]
    assert (
        len(layer_list[1].events.data.callbacks) == 1
    )  # `events` Event Emitter
    assert layer_list[1].source.parent() is layer_list[0]


def test_hide_unselected_layers():
    layer_list = make_three_layer_layerlist()
    layer_list[0].visible = True
    layer_list[1].visible = True
    layer_list[2].visible = True

    layer_list.selection.active = layer_list[1]

    assert layer_list[0].visible is True
    assert layer_list[1].visible is True
    assert layer_list[2].visible is True

    _hide_unselected(layer_list)

    assert layer_list[0].visible is False
    assert layer_list[1].visible is True
    assert layer_list[2].visible is False


def test_show_unselected_layers():
    layer_list = make_three_layer_layerlist()
    layer_list[0].visible = False
    layer_list[1].visible = True
    layer_list[2].visible = True

    layer_list.selection.active = layer_list[1]

    assert layer_list[0].visible is False
    assert layer_list[1].visible is True
    assert layer_list[2].visible is True

    _show_unselected(layer_list)

    assert layer_list[0].visible is True
    assert layer_list[1].visible is True
    assert layer_list[2].visible is True


def test_hide_selected_layers():
    layer_list = make_three_layer_layerlist()
    layer_list[0].visible = False
    layer_list[1].visible = True
    layer_list[2].visible = True

    layer_list.selection.active = layer_list[0]
    layer_list.selection.add(layer_list[1])

    assert layer_list[0].visible is False
    assert layer_list[1].visible is True
    assert layer_list[2].visible is True

    _hide_selected(layer_list)

    assert layer_list[0].visible is False
    assert layer_list[1].visible is False
    assert layer_list[2].visible is True


def test_show_selected_layers():
    layer_list = make_three_layer_layerlist()
    layer_list[0].visible = False
    layer_list[1].visible = True
    layer_list[2].visible = True

    layer_list.selection.active = layer_list[0]
    layer_list.selection.add(layer_list[1])

    assert layer_list[0].visible is False
    assert layer_list[1].visible is True
    assert layer_list[2].visible is True

    _show_selected(layer_list)

    assert layer_list[0].visible is True
    assert layer_list[1].visible is True
    assert layer_list[2].visible is True


@pytest.mark.parametrize(
    'mode', ['max', 'min', 'std', 'sum', 'mean', 'median']
)
def test_projections(mode):
    ll = LayerList()
    ll.append(
        Image(
            np.random.rand(7, 8, 8),
            scale=(3, 2, 2),
            translate=(10, 5, 5),
            units=('nm', 'um', 'um'),
            axis_labels=('z', 'y', 'x'),
        )
    )
    assert len(ll) == 1
    assert ll[-1].data.ndim == 3
    _project(ll, mode=mode)
    assert len(ll) == 2
    # because keepdims = False
    assert ll[-1].data.shape == (8, 8)
    assert tuple(ll[-1].scale) == (2, 2)
    assert tuple(ll[-1].translate) == (5, 5)
    assert ll[-1].units == (REG.um, REG.um)
    assert ll[-1].axis_labels == ('y', 'x')


@pytest.mark.parametrize(
    'mode',
    ['int8', 'int16', 'int32', 'int64', 'uint8', 'uint16', 'uint32', 'uint64'],
)
def test_convert_dtype(mode):
    ll = LayerList()
    data = np.zeros((10, 10), dtype=np.int16)
    ll.append(Labels(data))
    assert ll[-1].data.dtype == np.int16

    data[5, 5] = 1000
    assert data[5, 5] == 1000
    if mode == 'int8' or mode == 'uint8':
        # label value 1000 is outside of the target data type range.
        with pytest.raises(AssertionError):
            _convert_dtype(ll, mode=mode)
        assert ll[-1].data.dtype == np.int16
    else:
        _convert_dtype(ll, mode=mode)
        assert ll[-1].data.dtype == np.dtype(mode)

    assert ll[-1].data[5, 5] == 1000
    assert ll[-1].data.flatten().sum() == 1000


@pytest.mark.parametrize(
    ('layer', 'type_'),
    [
        (Image(np.random.rand(10, 10)), 'labels'),
        (Image(np.array([[1.5, 2.5], [3.5, 4.5]])), 'labels'),
        (Image(np.array([[1, 2], [3, 4]], dtype=(int))), 'labels'),
        (
            Image(zarr.array([[1, 2], [3, 4]], dtype=(int), chunks=(1, 2))),
            'labels',
        ),
        (Labels(np.ones((10, 10), dtype=int)), 'image'),
        (Shapes([np.array([[0, 0], [0, 10], [10, 0], [10, 10]])]), 'labels'),
    ],
)
def test_convert_layer(layer, type_):
    ll = LayerList()
    layer.scale *= 1.5
    original_scale = layer.scale.copy()
    ll.append(layer)
    assert ll[0]._type_string != type_
    _convert(ll, type_)
    if isinstance(layer, Shapes) or (
        type_ == 'labels'
        and isinstance(layer, Image)
        and np.issubdtype(layer.data.dtype, float)
    ):
        assert ll[1]._type_string == type_
        assert np.array_equal(ll[1].scale, original_scale)
    else:
        assert (
            layer.data is ll[0].data
        )  # check array data not copied unnecessarily


def test_convert_warns_with_projecton_mode():
    # inplace
    ll = LayerList(
        [Image(np.random.rand(10, 10).astype(int), projection_mode='mean')]
    )
    with pytest.warns(UserWarning, match='projection mode'):
        _convert(ll, 'labels')
    assert isinstance(ll['Image'], Labels)
    # not inplace
    ll = LayerList([Image(np.random.rand(10, 10), projection_mode='mean')])
    with pytest.warns(UserWarning, match='projection mode'):
        _convert(ll, 'labels')
    assert isinstance(ll['Image [1]'], Labels)


def make_three_layer_layerlist():
    layer_list = LayerList()
    layer_list.append(Points([[0, 0]], name='test'))
    layer_list.append(Image(np.random.rand(8, 8, 8)))
    layer_list.append(Image(np.random.rand(8, 8, 8)))

    return layer_list



import numpy as np
from scipy import ndimage as ndi

from napari.components.dims import Dims
from napari.layers import Labels
from napari.utils._proxies import ReadOnlyWrapper
from napari.utils.interactions import (
    mouse_move_callbacks,
    mouse_press_callbacks,
    mouse_release_callbacks,
)


def test_paint(MouseEvent):
    """Test painting labels with circle brush."""
    data = np.ones((20, 20), dtype=np.int32)
    layer = Labels(data)
    layer.brush_size = 10
    assert layer.cursor_size == 10

    layer.mode = 'paint'
    layer.selected_label = 3

    # Simulate click
    event = ReadOnlyWrapper(
        MouseEvent(
            type='mouse_press',
            is_dragging=False,
            position=(0, 0),
            view_direction=None,
            dims_displayed=(0, 1),
            dims_point=(0, 0),
        )
    )
    mouse_press_callbacks(layer, event)

    # Simulate drag
    event = ReadOnlyWrapper(
        MouseEvent(
            type='mouse_move',
            is_dragging=True,
            position=(19, 19),
            view_direction=None,
            dims_displayed=(0, 1),
            dims_point=(0, 0),
        )
    )
    mouse_move_callbacks(layer, event)

    # Simulate release
    event = ReadOnlyWrapper(
        MouseEvent(
            type='mouse_release',
            is_dragging=False,
            position=(19, 19),
            view_direction=None,
            dims_displayed=(0, 1),
            dims_point=(0, 0),
        )
    )
    mouse_release_callbacks(layer, event)

    # Painting goes from (0, 0) to (19, 19) with a brush size of 10, changing
    # all pixels along that path, but none outside it.
    assert np.unique(layer.data[:8, :8]) == 3
    assert np.unique(layer.data[-8:, -8:]) == 3
    assert np.unique(layer.data[:5, -5:]) == 1
    assert np.unique(layer.data[-5:, :5]) == 1
    assert np.sum(layer.data == 3) == 244


def test_paint_scale(MouseEvent):
    """Test painting labels with circle brush when scaled."""
    data = np.ones((20, 20), dtype=np.int32)
    layer = Labels(data, scale=(2, 2))
    layer.brush_size = 10

    layer.mode = 'paint'
    layer.selected_label = 3

    # Simulate click
    event = ReadOnlyWrapper(
        MouseEvent(
            type='mouse_press',
            is_dragging=False,
            position=(0, 0),
            view_direction=None,
            dims_displayed=(0, 1),
            dims_point=(0, 0),
        )
    )
    mouse_press_callbacks(layer, event)

    # Simulate drag
    event = ReadOnlyWrapper(
        MouseEvent(
            type='mouse_move',
            is_dragging=True,
            position=(39, 39),
            view_direction=None,
            dims_displayed=(0, 1),
            dims_point=(0, 0),
        )
    )
    mouse_move_callbacks(layer, event)

    # Simulate release
    event = ReadOnlyWrapper(
        MouseEvent(
            type='mouse_release',
            is_dragging=False,
            position=(39, 39),
            view_direction=None,
            dims_displayed=(0, 1),
            dims_point=(0, 0),
        )
    )
    mouse_release_callbacks(layer, event)

    # Painting goes from (0, 0) to (19, 19) with a brush size of 10, changing
    # all pixels along that path, but none outside it.
    assert np.unique(layer.data[:8, :8]) == 3
    assert np.unique(layer.data[-8:, -8:]) == 3
    assert np.unique(layer.data[:5, -5:]) == 1
    assert np.unique(layer.data[-5:, :5]) == 1
    assert np.sum(layer.data == 3) == 244


def test_erase(MouseEvent):
    """Test erasing labels with different brush shapes."""
    data = np.ones((20, 20), dtype=np.int32)
    layer = Labels(data)
    layer.brush_size = 10
    layer.mode = 'erase'

    layer.selected_label = 3

    # Simulate click
    event = ReadOnlyWrapper(
        MouseEvent(
            type='mouse_press',
            is_dragging=False,
            position=(0, 0),
            view_direction=None,
            dims_displayed=(0, 1),
            dims_point=(0, 0),
        )
    )
    mouse_press_callbacks(layer, event)

    # Simulate drag
    event = ReadOnlyWrapper(
        MouseEvent(
            type='mouse_move',
            is_dragging=True,
            position=(19, 19),
            view_direction=None,
            dims_displayed=(0, 1),
            dims_point=(0, 0),
        )
    )
    mouse_move_callbacks(layer, event)

    # Simulate release
    event = ReadOnlyWrapper(
        MouseEvent(
            type='mouse_release',
            is_dragging=False,
            position=(19, 19),
            view_direction=None,
            dims_displayed=(0, 1),
            dims_point=(0, 0),
        )
    )
    mouse_release_callbacks(layer, event)

    # Painting goes from (0, 0) to (19, 19) with a brush size of 10, changing
    # all pixels along that path, but non outside it.
    assert np.unique(layer.data[:8, :8]) == 0
    assert np.unique(layer.data[-8:, -8:]) == 0
    assert np.unique(layer.data[:5, -5:]) == 1
    assert np.unique(layer.data[-5:, :5]) == 1
    assert np.sum(layer.data == 1) == 156


def test_pick(MouseEvent):
    """Test picking label."""
    data = np.ones((20, 20), dtype=np.int32)
    data[:5, :5] = 2
    data[-5:, -5:] = 3
    layer = Labels(data)
    assert layer.selected_label == 1

    layer.mode = 'pick'

    # Simulate click
    event = ReadOnlyWrapper(
        MouseEvent(
            type='mouse_press',
            is_dragging=False,
            position=(0, 0),
            view_direction=None,
            dims_displayed=(0, 1),
            dims_point=(0, 0),
        )
    )
    mouse_press_callbacks(layer, event)
    assert layer.selected_label == 2

    # Simulate click
    event = ReadOnlyWrapper(
        MouseEvent(
            type='mouse_press',
            is_dragging=False,
            position=(19, 19),
            view_direction=None,
            dims_displayed=(0, 1),
            dims_point=(0, 0),
        )
    )
    mouse_press_callbacks(layer, event)
    assert layer.selected_label == 3


def test_fill(MouseEvent):
    """Test filling label."""
    data = np.ones((20, 20), dtype=np.int32)
    data[:5, :5] = 2
    data[-5:, -5:] = 3
    layer = Labels(data)
    assert np.unique(layer.data[:5, :5]) == 2
    assert np.unique(layer.data[-5:, -5:]) == 3
    assert np.unique(layer.data[:5, -5:]) == 1
    assert np.unique(layer.data[-5:, :5]) == 1

    layer.mode = 'fill'
    layer.selected_label = 4

    # Simulate click
    event = ReadOnlyWrapper(
        MouseEvent(
            type='mouse_press',
            is_dragging=False,
            position=(0, 0),
            view_direction=None,
            dims_displayed=(0, 1),
            dims_point=(0, 0),
        )
    )
    mouse_press_callbacks(layer, event)
    assert np.unique(layer.data[:5, :5]) == 4
    assert np.unique(layer.data[-5:, -5:]) == 3
    assert np.unique(layer.data[:5, -5:]) == 1
    assert np.unique(layer.data[-5:, :5]) == 1

    layer.selected_label = 5

    # Simulate click
    event = ReadOnlyWrapper(
        MouseEvent(
            type='mouse_press',
            is_dragging=False,
            position=(19, 19),
            view_direction=None,
            dims_displayed=(0, 1),
            dims_point=(0, 0),
        )
    )
    mouse_press_callbacks(layer, event)
    assert np.unique(layer.data[:5, :5]) == 4
    assert np.unique(layer.data[-5:, -5:]) == 5
    assert np.unique(layer.data[:5, -5:]) == 1
    assert np.unique(layer.data[-5:, :5]) == 1


def test_fill_nD_plane(MouseEvent):
    """Test filling label nD plane."""
    data = np.ones((20, 20, 20), dtype=np.int32)
    data[:5, :5, :5] = 2
    data[0, 8:10, 8:10] = 2
    data[-5:, -5:, -5:] = 3
    layer = Labels(data)
    assert np.unique(layer.data[:5, :5, :5]) == 2
    assert np.unique(layer.data[-5:, -5:, -5:]) == 3
    assert np.unique(layer.data[:5, -5:, -5:]) == 1
    assert np.unique(layer.data[-5:, :5, -5:]) == 1
    assert np.unique(layer.data[0, 8:10, 8:10]) == 2

    layer.mode = 'fill'
    layer.selected_label = 4

    # Simulate click
    event = ReadOnlyWrapper(
        MouseEvent(
            type='mouse_press',
            is_dragging=False,
            position=(0, 0, 0),
            view_direction=(1, 0, 0),
            dims_displayed=(0, 1),
            dims_point=(0, 0),
        )
    )
    mouse_press_callbacks(layer, event)
    assert np.unique(layer.data[0, :5, :5]) == 4
    assert np.unique(layer.data[1:5, :5, :5]) == 2
    assert np.unique(layer.data[-5:, -5:, -5:]) == 3
    assert np.unique(layer.data[:5, -5:, -5:]) == 1
    assert np.unique(layer.data[-5:, :5, -5:]) == 1
    assert np.unique(layer.data[0, 8:10, 8:10]) == 2

    layer.selected_label = 5

    # Simulate click
    event = ReadOnlyWrapper(
        MouseEvent(
            type='mouse_press',
            is_dragging=False,
            position=(0, 19, 19),
            view_direction=(1, 0, 0),
            dims_displayed=(0, 1),
            dims_point=(0, 0, 0),
        )
    )
    mouse_press_callbacks(layer, event)
    assert np.unique(layer.data[0, :5, :5]) == 4
    assert np.unique(layer.data[1:5, :5, :5]) == 2
    assert np.unique(layer.data[-5:, -5:, -5:]) == 3
    assert np.unique(layer.data[1:5, -5:, -5:]) == 1
    assert np.unique(layer.data[-5:, :5, -5:]) == 1
    assert np.unique(layer.data[0, -5:, -5:]) == 5
    assert np.unique(layer.data[0, :5, -5:]) == 5
    assert np.unique(layer.data[0, 8:10, 8:10]) == 2


def test_fill_nD_all(MouseEvent):
    """Test filling label nD."""
    data = np.ones((20, 20, 20), dtype=np.int32)
    data[:5, :5, :5] = 2
    data[0, 8:10, 8:10] = 2
    data[-5:, -5:, -5:] = 3
    layer = Labels(data)
    assert np.unique(layer.data[:5, :5, :5]) == 2
    assert np.unique(layer.data[-5:, -5:, -5:]) == 3
    assert np.unique(layer.data[:5, -5:, -5:]) == 1
    assert np.unique(layer.data[-5:, :5, -5:]) == 1
    assert np.unique(layer.data[0, 8:10, 8:10]) == 2

    layer.n_edit_dimensions = 3
    layer.mode = 'fill'
    layer.selected_label = 4

    # Simulate click
    event = ReadOnlyWrapper(
        MouseEvent(
            type='mouse_press',
            is_dragging=False,
            position=(0, 0, 0),
            view_direction=(1, 0, 0),
            dims_displayed=(0, 1),
            dims_point=(0, 0),
        )
    )
    mouse_press_callbacks(layer, event)
    assert np.unique(layer.data[:5, :5, :5]) == 4
    assert np.unique(layer.data[-5:, -5:, -5:]) == 3
    assert np.unique(layer.data[:5, -5:, -5:]) == 1
    assert np.unique(layer.data[-5:, :5, -5:]) == 1
    assert np.unique(layer.data[0, 8:10, 8:10]) == 2

    layer.selected_label = 5

    # Simulate click
    event = ReadOnlyWrapper(
        MouseEvent(
            type='mouse_press',
            is_dragging=False,
            position=(0, 19, 19),
            view_direction=(1, 0, 0),
            dims_displayed=(0, 1),
            dims_point=(0, 0, 0),
        )
    )
    mouse_press_callbacks(layer, event)
    assert np.unique(layer.data[:5, :5, :5]) == 4
    assert np.unique(layer.data[-5:, -5:, -5:]) == 3
    assert np.unique(layer.data[:5, -5:, -5:]) == 5
    assert np.unique(layer.data[-5:, :5, -5:]) == 5
    assert np.unique(layer.data[0, 8:10, 8:10]) == 2


def test_paint_3d(MouseEvent):
    """Test filling label nD."""
    data = np.zeros((21, 21, 21), dtype=np.int32)
    data[10, 10, 10] = 1
    layer = Labels(data)
    layer._slice_dims(Dims(ndim=3, ndisplay=3))

    layer.n_edit_dimensions = 3
    layer.mode = 'paint'
    layer.selected_label = 4
    layer.brush_size = 3

    # Simulate click
    event = ReadOnlyWrapper(
        MouseEvent(
            type='mouse_press',
            is_dragging=False,
            position=(0.1, 0, 0),
            view_direction=np.full(3, np.sqrt(3)),
            dims_displayed=(0, 1, 2),
            dims_point=(0, 0, 0),
        )
    )
    mouse_press_callbacks(layer, event)
    np.testing.assert_array_equal(np.unique(layer.data), [0, 4])
    num_filled = np.bincount(layer.data.ravel())[4]
    assert num_filled > 1

    layer.mode = 'erase'

    # Simulate click
    event = ReadOnlyWrapper(
        MouseEvent(
            type='mouse_press',
            is_dragging=False,
            position=(0, 10, 10),
            view_direction=(1, 0, 0),
            dims_displayed=(0, 1, 2),
            dims_point=(0, 0, 0),
        )
    )
    mouse_press_callbacks(layer, event)

    new_num_filled = np.bincount(layer.data.ravel())[4]
    assert new_num_filled < num_filled


def test_erase_3d_undo(MouseEvent):
    """Test erasing labels in 3D then undoing the erase.

    Specifically, this test checks that undo is correctly filled even
    when a click and drag starts outside of the data volume.
    """
    data = np.zeros((20, 20, 20), dtype=np.int32)
    data[10, :, :] = 1
    layer = Labels(data)
    layer.brush_size = 5
    layer.mode = 'erase'
    layer._slice_dims(Dims(ndim=3, ndisplay=3))
    layer.n_edit_dimensions = 3

    # Simulate click
    event = ReadOnlyWrapper(
        MouseEvent(
            type='mouse_press',
            is_dragging=False,
            position=(-1, -1, -1),
            view_direction=(1, 0, 0),
            dims_displayed=(0, 1, 2),
            dims_point=(0, 0, 0),
        )
    )
    mouse_press_callbacks(layer, event)

    # Simulate drag. Note: we need to include top left and bottom right in the
    # drag or there are no coordinates to interpolate
    event = ReadOnlyWrapper(
        MouseEvent(
            type='mouse_move',
            is_dragging=True,
            position=(-1, 0.1, 0.1),
            view_direction=(1, 0, 0),
            dims_displayed=(0, 1, 2),
            dims_point=(0, 0, 0),
        )
    )
    mouse_move_callbacks(layer, event)
    event = ReadOnlyWrapper(
        MouseEvent(
            type='mouse_move',
            is_dragging=True,
            position=(-1, 18.9, 18.9),
            view_direction=(1, 0, 0),
            dims_displayed=(0, 1, 2),
            dims_point=(0, 0, 0),
        )
    )
    mouse_move_callbacks(layer, event)

    # Simulate release
    event = ReadOnlyWrapper(
        MouseEvent(
            type='mouse_release',
            is_dragging=False,
            position=(-1, 21, 21),
            view_direction=(1, 0, 0),
            dims_displayed=(0, 1, 2),
            dims_point=(0, 0, 0),
        )
    )
    mouse_release_callbacks(layer, event)

    # Erasing goes from (-1, -1, -1) to (-1, 21, 21), should split the labels
    # into two sections. Undoing should work and reunite the labels to one
    # square
    assert ndi.label(layer.data)[1] == 2
    layer.undo()
    assert ndi.label(layer.data)[1] == 1


def test_erase_3d_undo_empty(MouseEvent):
    """Nothing should be added to undo queue when clicks fall outside data."""
    data = np.zeros((20, 20, 20), dtype=np.int32)
    data[10, :, :] = 1
    layer = Labels(data)
    layer.brush_size = 5
    layer.mode = 'erase'
    layer._slice_dims(Dims(ndim=3, ndisplay=3))
    layer.n_edit_dimensions = 3

    # Simulate click, outside data
    event = ReadOnlyWrapper(
        MouseEvent(
            type='mouse_press',
            is_dragging=False,
            position=(-1, -1, -1),
            view_direction=(1, 0, 0),
            dims_displayed=(0, 1, 2),
            dims_point=(0, 0, 0),
        )
    )
    mouse_press_callbacks(layer, event)

    # Simulate release
    event = ReadOnlyWrapper(
        MouseEvent(
            type='mouse_release',
            is_dragging=False,
            position=(-1, -1, -1),
            view_direction=(1, 0, 0),
            dims_displayed=(0, 1, 2),
            dims_point=(0, 0, 0),
        )
    )
    mouse_release_callbacks(layer, event)

    # Undo queue should be empty
    assert len(layer._undo_history) == 0



from __future__ import annotations

from enum import Enum

from qtpy.QtCore import QPoint, QSize, Qt
from qtpy.QtGui import QCursor, QPainter, QPen, QPixmap


def crosshair_pixmap():
    """Create a cross cursor with white/black hollow square pixmap in the middle.
    For use as points cursor."""

    size = 25

    pixmap = QPixmap(QSize(size, size))
    pixmap.fill(Qt.GlobalColor.transparent)
    painter = QPainter(pixmap)

    # Base measures
    width = 1
    center = 3  # Must be odd!
    rect_size = center + 2 * width
    square = rect_size + width * 4

    pen = QPen(Qt.GlobalColor.white, 1)
    pen.setJoinStyle(Qt.PenJoinStyle.MiterJoin)
    painter.setPen(pen)

    # # Horizontal rectangle
    painter.drawRect(0, (size - rect_size) // 2, size - 1, rect_size - 1)

    # Vertical rectangle
    painter.drawRect((size - rect_size) // 2, 0, rect_size - 1, size - 1)

    # Square
    painter.drawRect(
        (size - square) // 2, (size - square) // 2, square - 1, square - 1
    )

    pen = QPen(Qt.GlobalColor.black, 2)
    pen.setJoinStyle(Qt.PenJoinStyle.MiterJoin)
    painter.setPen(pen)

    # # Square
    painter.drawRect(
        (size - square) // 2 + 2,
        (size - square) // 2 + 2,
        square - 4,
        square - 4,
    )

    pen = QPen(Qt.GlobalColor.black, 3)
    pen.setJoinStyle(Qt.PenJoinStyle.MiterJoin)
    painter.setPen(pen)

    # # # Horizontal lines
    mid_vpoint = QPoint(2, size // 2)
    painter.drawLine(
        mid_vpoint, QPoint(((size - center) // 2) - center + 1, size // 2)
    )
    mid_vpoint = QPoint(size - 3, size // 2)
    painter.drawLine(
        mid_vpoint, QPoint(((size - center) // 2) + center + 1, size // 2)
    )

    # # # Vertical lines
    mid_hpoint = QPoint(size // 2, 2)
    painter.drawLine(
        QPoint(size // 2, ((size - center) // 2) - center + 1), mid_hpoint
    )
    mid_hpoint = QPoint(size // 2, size - 3)
    painter.drawLine(
        QPoint(size // 2, ((size - center) // 2) + center + 1), mid_hpoint
    )

    painter.end()
    return pixmap


def square_pixmap(size: int):
    """Create a white/black hollow square pixmap. For use as labels cursor."""
    size = max(int(size), 1)
    pixmap = QPixmap(QSize(size, size))
    pixmap.fill(Qt.GlobalColor.transparent)
    painter = QPainter(pixmap)
    painter.setPen(Qt.GlobalColor.white)
    painter.drawRect(0, 0, size - 1, size - 1)
    painter.setPen(Qt.GlobalColor.black)
    painter.drawRect(1, 1, size - 3, size - 3)
    painter.end()
    return pixmap


def create_square_cursor(size: int):
    return QCursor(square_pixmap(size))


def create_crosshair_cursor():
    return QCursor(crosshair_pixmap())


def create_blank_cursor():
    return QCursor(Qt.BlankCursor)


class QtCursorVisual(Enum):
    blank = staticmethod(create_blank_cursor)
    square = staticmethod(create_square_cursor)
    cross = Qt.CursorShape.CrossCursor
    forbidden = Qt.CursorShape.ForbiddenCursor
    pointing = Qt.CursorShape.PointingHandCursor
    standard = Qt.CursorShape.ArrowCursor
    crosshair = staticmethod(create_crosshair_cursor)



import pytest

from napari.utils.migrations import (
    _DeprecatingDict,
    add_deprecated_property,
    deprecated_class_name,
    rename_argument,
)


def test_simple():
    @rename_argument('a', 'b', '1', '0.5')
    def sample_fun(b):
        return b

    assert sample_fun(1) == 1
    assert sample_fun(b=1) == 1
    with pytest.deprecated_call():
        assert sample_fun(a=1) == 1
    with pytest.raises(ValueError, match='already defined'):
        sample_fun(b=1, a=1)


def test_constructor():
    class Sample:
        @rename_argument('a', 'b', '1', '0.5')
        def __init__(self, b) -> None:
            self.b = b

    assert Sample(1).b == 1
    assert Sample(b=1).b == 1
    with pytest.deprecated_call():
        assert Sample(a=1).b == 1


def test_deprecated_property() -> None:
    class Dummy:
        def __init__(self) -> None:
            self._value = 0

        @property
        def new_property(self) -> int:
            return self._value

        @new_property.setter
        def new_property(self, value: int) -> int:
            self._value = value

    instance = Dummy()

    add_deprecated_property(
        Dummy, 'old_property', 'new_property', '0.1.0', '0.0.0'
    )

    assert instance.new_property == 0

    instance.new_property = 1

    msg = 'Dummy.old_property is deprecated since 0.0.0 and will be removed in 0.1.0. Please use new_property'

    with pytest.warns(FutureWarning, match=msg):
        assert instance.old_property == 1

    with pytest.warns(FutureWarning, match=msg):
        instance.old_property = 2

    assert instance.new_property == 2


def test_deprecated_class_name():
    """Test the deprecated class name function."""

    class macOS:
        pass

    MacOSX = deprecated_class_name(
        macOS, 'MacOSX', version='10.12', since_version='10.11'
    )

    with pytest.warns(FutureWarning, match='deprecated.*macOS'):
        _os = MacOSX()

    with pytest.warns(FutureWarning, match='deprecated.*macOS'):

        class MacOSXServer(MacOSX):
            pass


def test_deprecating_dict_with_renamed_in_deprecated_keys():
    d = _DeprecatingDict({'a': 1, 'b': 2})
    d.set_deprecated_from_rename(
        from_name='c', to_name='a', version='v2.0', since_version='v1.6'
    )
    assert 'c' in d.deprecated_keys


def test_deprecating_dict_with_renamed_getitem_deprecated():
    d = _DeprecatingDict({'a': 1, 'b': 2})
    d.set_deprecated_from_rename(
        from_name='c', to_name='a', version='v2.0', since_version='v1.6'
    )
    with pytest.warns(FutureWarning, match='is deprecated since'):
        assert d['c'] == 1


def test_deprecating_dict_with_renamed_get_deprecated():
    d = _DeprecatingDict({'a': 1, 'b': 2})
    d.set_deprecated_from_rename(
        from_name='c', to_name='a', version='v2.0', since_version='v1.6'
    )
    with pytest.warns(FutureWarning, match='is deprecated since'):
        assert d.get('c') == 1


def test_deprecating_dict_with_renamed_set_nondeprecated():
    d = _DeprecatingDict({'a': 1, 'b': 2})
    d.set_deprecated_from_rename(
        from_name='c', to_name='a', version='v2.0', since_version='v1.6'
    )

    d['a'] = 3

    with pytest.warns(FutureWarning, match='is deprecated since'):
        assert d['c'] == 3


def test_deprecating_dict_with_renamed_set_deprecated():
    d = _DeprecatingDict({'a': 1, 'b': 2})
    d.set_deprecated_from_rename(
        from_name='c', to_name='a', version='v2.0', since_version='v1.6'
    )

    with pytest.warns(FutureWarning, match='is deprecated since'):
        d['c'] = 3

    with pytest.warns(FutureWarning, match='is deprecated since'):
        assert d['c'] == 3
    assert d['a'] == 3


def test_deprecating_dict_with_renamed_update_nondeprecated():
    d = _DeprecatingDict({'a': 1, 'b': 2})
    d.set_deprecated_from_rename(
        from_name='c', to_name='a', version='v2.0', since_version='v1.6'
    )

    d.update({'a': 3})

    with pytest.warns(FutureWarning, match='is deprecated since'):
        assert d['c'] == 3


def test_deprecating_dict_with_renamed_update_deprecated():
    d = _DeprecatingDict({'a': 1, 'b': 2})
    d.set_deprecated_from_rename(
        from_name='c', to_name='a', version='v2.0', since_version='v1.6'
    )

    with pytest.warns(FutureWarning, match='is deprecated since'):
        d.update({'c': 3})

    with pytest.warns(FutureWarning, match='is deprecated since'):
        assert d['c'] == 3
    assert d['a'] == 3


def test_deprecating_dict_with_renamed_del_nondeprecated():
    d = _DeprecatingDict({'a': 1, 'b': 2})
    d.set_deprecated_from_rename(
        from_name='c', to_name='a', version='v2.0', since_version='v1.6'
    )
    assert 'a' in d
    with pytest.warns(FutureWarning, match='is deprecated since'):
        assert 'c' in d

    with pytest.warns(FutureWarning, match='is deprecated since'):
        del d['c']

    assert 'a' not in d
    with pytest.warns(FutureWarning, match='is deprecated since'):
        assert 'c' not in d


def test_deprecating_dict_with_renamed_del_deprecated():
    d = _DeprecatingDict({'a': 1, 'b': 2})
    d.set_deprecated_from_rename(
        from_name='c', to_name='a', version='v2.0', since_version='v1.6'
    )
    with pytest.warns(FutureWarning, match='is deprecated since'):
        assert 'c' in d
    assert 'a' in d

    with pytest.warns(FutureWarning, match='is deprecated since'):
        del d['c']

    with pytest.warns(FutureWarning, match='is deprecated since'):
        assert 'c' not in d
    assert 'a' not in d


def test_deprecating_dict_with_renamed_pop_nondeprecated():
    d = _DeprecatingDict({'a': 1, 'b': 2})
    d.set_deprecated_from_rename(
        from_name='c', to_name='a', version='v2.0', since_version='v1.6'
    )
    assert 'a' in d
    with pytest.warns(FutureWarning, match='is deprecated since'):
        assert 'c' in d

    with pytest.warns(FutureWarning, match='is deprecated since'):
        d.pop('c')

    assert 'a' not in d
    with pytest.warns(FutureWarning, match='is deprecated since'):
        assert 'c' not in d


def test_deprecating_dict_with_renamed_pop_deprecated():
    d = _DeprecatingDict({'a': 1, 'b': 2})
    d.set_deprecated_from_rename(
        from_name='c', to_name='a', version='v2.0', since_version='v1.6'
    )
    with pytest.warns(FutureWarning, match='is deprecated since'):
        assert 'c' in d
    assert 'a' in d

    with pytest.warns(FutureWarning, match='is deprecated since'):
        d.pop('c')

    with pytest.warns(FutureWarning, match='is deprecated since'):
        assert 'c' not in d
    assert 'a' not in d


def test_deprecating_dict_with_renamed_copy():
    d = _DeprecatingDict({'a': 1, 'b': 2})
    d.set_deprecated_from_rename(
        from_name='c', to_name='a', version='v2.0', since_version='v1.6'
    )

    e = d.copy()

    assert d is not e
    assert e.data == d.data
    assert e.deprecated_keys == d.deprecated_keys



from vispy.scene.visuals import Compound, Ellipse

from napari._vispy.overlays.base import ViewerOverlayMixin, VispyCanvasOverlay


class VispyBrushCircleOverlay(ViewerOverlayMixin, VispyCanvasOverlay):
    def __init__(self, *, viewer, overlay, parent=None):
        self._white_circle = Ellipse(
            center=(0, 0),
            color=(0, 0, 0, 0.0),
            border_color='white',
            border_method='agg',
        )
        self._black_circle = Ellipse(
            center=(0, 0),
            color=(0, 0, 0, 0.0),
            border_color='black',
            border_method='agg',
        )

        super().__init__(
            node=Compound([self._white_circle, self._black_circle]),
            viewer=viewer,
            overlay=overlay,
            parent=parent,
        )

        self._last_mouse_pos = None

        self.overlay.events.size.connect(self._on_size_change)
        self.node.events.canvas_change.connect(self._on_canvas_change)
        self.viewer.events.mouse_over_canvas.connect(
            self._on_mouse_over_canvas
        )
        # no need to connect position, since that's in the base classes of CanvasOverlay

        self.reset()

    def _on_position_change(self, event=None):
        self._set_position(self.overlay.position)

    def _on_size_change(self, event=None):
        self._white_circle.radius = self.overlay.size / 2
        self._black_circle.radius = self._white_circle.radius - 1

    def _on_visible_change(self):
        if self._last_mouse_pos is not None:
            self._set_position(self._last_mouse_pos)
        self.node.visible = (
            self.overlay.visible and self.viewer.mouse_over_canvas
        )

    def _on_mouse_move(self, event):
        self._last_mouse_pos = event.pos
        if self.overlay.visible:
            self.overlay.position = event.pos.tolist()

    def _set_position(self, pos):
        if not self.overlay.position_is_frozen:
            self.node.transform.translate = [pos[0], pos[1], 0, 0]

    def _on_canvas_change(self, event):
        if event.new is not None:
            event.new.events.mouse_move.connect(self._on_mouse_move)
        if event.old is not None:
            event.old.events.mouse_move.disconnect(self._on_mouse_move)

    def _on_mouse_over_canvas(self):
        if self.viewer.mouse_over_canvas:
            # Move the cursor outside the canvas when the mouse leaves it.
            # It fixes the bug described in PR #5763:
            # https://github.com/napari/napari/pull/5763#issuecomment-1523182141
            self._set_position((-1000, -1000))
            self.node.visible = self.overlay.visible
        else:
            if self.overlay.visible:
                self.node.visible = self.overlay.position_is_frozen
            else:
                self.node.visible = False

    def reset(self):
        super().reset()
        self._on_size_change()
        self._last_mouse_pos = None



import importlib
from unittest import mock

import pytest
from app_model.types import MenuItem, SubmenuItem
from npe2 import DynamicPlugin
from qtpy.QtWidgets import QWidget

from napari._app_model import get_app_model
from napari._app_model.constants import MenuId
from napari._qt._qapp_model.qactions import _plugins, init_qactions
from napari._qt._qplugins._qnpe2 import _toggle_or_get_widget
from napari._tests.utils import skip_local_popups
from napari.plugins._tests.test_npe2 import mock_pm  # noqa: F401


class DummyWidget(QWidget):
    pass


@pytest.mark.skipif(
    not _plugins._plugin_manager_dialog_avail(),
    reason='`napari_plugin_manager` not available',
)
def test_plugin_manager_action(make_napari_viewer):
    """
    Test manage plugins installation action.

    The test is skipped in case `napari_plugin_manager` is not available
    """
    app = get_app_model()
    viewer = make_napari_viewer()

    with mock.patch(
        'napari_plugin_manager.qt_plugin_dialog.QtPluginDialog'
    ) as mock_plugin_dialog:
        app.commands.execute_command(
            'napari.window.plugins.plugin_install_dialog'
        )
    mock_plugin_dialog.assert_called_once_with(viewer.window._qt_window)


def test_plugin_errors_action(make_napari_viewer):
    """Test plugin errors action."""
    make_napari_viewer()
    app = get_app_model()

    with mock.patch(
        'napari._qt._qapp_model.qactions._plugins.QtPluginErrReporter.exec_'
    ) as mock_plugin_dialog:
        app.commands.execute_command(
            'napari.window.plugins.plugin_err_reporter'
        )
    mock_plugin_dialog.assert_called_once()


@skip_local_popups
def test_toggle_or_get_widget(
    make_napari_viewer,
    tmp_plugin: DynamicPlugin,
    qtbot,
):
    """Check `_toggle_or_get_widget` changes visibility correctly."""
    widget_name = 'Widget'
    plugin_name = 'tmp_plugin'
    full_name = 'Widget (Temp Plugin)'

    @tmp_plugin.contribute.widget(display_name=widget_name)
    def widget1():
        return DummyWidget()

    app = get_app_model()
    # Viewer needs to be visible
    viewer = make_napari_viewer(show=True)

    # Trigger the action, opening the widget: `Widget 1`
    app.commands.execute_command('tmp_plugin:Widget')
    widget = viewer.window._dock_widgets[full_name]
    # Widget takes some time to appear
    qtbot.waitUntil(widget.isVisible)
    assert widget.isVisible()

    # Define not visible callable
    def widget_not_visible():
        return not widget.isVisible()

    # Hide widget
    _toggle_or_get_widget(
        plugin=plugin_name,
        widget_name=widget_name,
        full_name=full_name,
    )
    qtbot.waitUntil(widget_not_visible)
    assert not widget.isVisible()

    # Make widget appear again
    _toggle_or_get_widget(
        plugin=plugin_name,
        widget_name=widget_name,
        full_name=full_name,
    )
    qtbot.waitUntil(widget.isVisible)
    assert widget.isVisible()


def test_plugin_single_widget_menu(
    make_napari_viewer, tmp_plugin: DynamicPlugin
):
    """Test single plugin widgets get added to the window menu correctly."""

    @tmp_plugin.contribute.widget(display_name='Widget 1')
    def widget1():
        return DummyWidget()

    app = get_app_model()
    viewer = make_napari_viewer()

    assert tmp_plugin.display_name == 'Temp Plugin'
    plugin_menu = app.menus.get_menu('napari/plugins')
    assert plugin_menu[0].command.title == 'Widget 1 (Temp Plugin)'
    assert len(viewer.window._dock_widgets) == 0
    assert 'tmp_plugin:Widget 1' in app.commands
    # trigger the action, opening the widget: `Widget 1`
    app.commands.execute_command('tmp_plugin:Widget 1')
    assert len(viewer.window._dock_widgets) == 1
    assert 'Widget 1 (Temp Plugin)' in viewer.window._dock_widgets


def test_plugin_multiple_widget_menu(
    make_napari_viewer,
    tmp_plugin: DynamicPlugin,
):
    """Check plugin with >1 widgets added with submenu and uses 'display_name'."""

    @tmp_plugin.contribute.widget(display_name='Widget 1')
    def widget1():
        return DummyWidget()

    @tmp_plugin.contribute.widget(display_name='Widget 2')
    def widget2():
        return DummyWidget()

    app = get_app_model()
    viewer = make_napari_viewer()

    assert tmp_plugin.display_name == 'Temp Plugin'
    plugin_menu = app.menus.get_menu('napari/plugins')
    assert plugin_menu[0].title == tmp_plugin.display_name
    plugin_submenu = app.menus.get_menu('napari/plugins/tmp_plugin')
    assert plugin_submenu[0].command.title == 'Widget 1'
    assert len(viewer.window._dock_widgets) == 0
    assert 'tmp_plugin:Widget 1' in app.commands
    # Trigger the action, opening the first widget: `Widget 1`
    app.commands.execute_command('tmp_plugin:Widget 1')
    assert len(viewer.window._dock_widgets) == 1
    assert 'Widget 1 (Temp Plugin)' in viewer.window._dock_widgets


def test_plugin_menu_plugin_state_change(
    make_napari_viewer,
    tmp_plugin: DynamicPlugin,
):
    """Check plugin menu items correct after a plugin changes state."""
    app = get_app_model()
    pm = tmp_plugin.plugin_manager

    # Register plugin q actions
    init_qactions()
    # Check only `Q_PLUGINS_ACTIONS` in plugin menu before any plugins registered
    plugins_menu = app.menus.get_menu(MenuId.MENUBAR_PLUGINS)
    assert len(plugins_menu) == len(_plugins.Q_PLUGINS_ACTIONS)

    @tmp_plugin.contribute.widget(display_name='Widget 1')
    def widget1():
        """Dummy widget."""

    @tmp_plugin.contribute.widget(display_name='Widget 2')
    def widget2():
        """Dummy widget."""

    # Configures `app`, registers actions and initializes plugins
    make_napari_viewer()

    plugins_menu = app.menus.get_menu(MenuId.MENUBAR_PLUGINS)
    assert len(plugins_menu) == len(_plugins.Q_PLUGINS_ACTIONS) + 1
    assert isinstance(plugins_menu[-1], SubmenuItem)
    assert plugins_menu[-1].title == tmp_plugin.display_name
    plugin_submenu = app.menus.get_menu(MenuId.MENUBAR_PLUGINS + '/tmp_plugin')
    assert len(plugin_submenu) == 2
    assert isinstance(plugin_submenu[0], MenuItem)
    assert plugin_submenu[0].command.title == 'Widget 1'
    assert 'tmp_plugin:Widget 1' in app.commands

    # Disable plugin
    pm.disable(tmp_plugin.name)
    with pytest.raises(KeyError):
        app.menus.get_menu(MenuId.MENUBAR_PLUGINS + '/tmp_plugin')
    assert 'tmp_plugin:Widget 1' not in app.commands

    # Enable plugin
    pm.enable(tmp_plugin.name)
    samples_sub_menu = app.menus.get_menu(
        MenuId.MENUBAR_PLUGINS + '/tmp_plugin'
    )
    assert len(samples_sub_menu) == 2
    assert 'tmp_plugin:Widget 1' in app.commands


def test_plugin_widget_checked(
    make_napari_viewer, qtbot, tmp_plugin: DynamicPlugin
):
    """Check widget toggling/hiding updates check mark correctly."""

    @tmp_plugin.contribute.widget(display_name='Widget')
    def widget_contrib():
        return DummyWidget()

    app = get_app_model()
    viewer = make_napari_viewer()

    assert 'tmp_plugin:Widget' in app.commands
    widget_action = viewer.window.plugins_menu.findAction('tmp_plugin:Widget')
    assert not widget_action.isChecked()
    # Trigger the action, opening the widget
    widget_action.trigger()
    assert widget_action.isChecked()
    assert 'Widget (Temp Plugin)' in viewer.window._dock_widgets
    widget = viewer.window._dock_widgets['Widget (Temp Plugin)']

    # Hide widget
    widget.title.hide_button.click()
    # Run `_on_about_to_show`, which is called on `aboutToShow`` event,
    # to update checked status
    viewer.window.plugins_menu._on_about_to_show()
    assert not widget_action.isChecked()

    # Trigger the action again to open widget and test item checked
    widget_action.trigger()
    assert widget_action.isChecked()


def test_import_plugin_manager():
    from napari_plugin_manager.qt_plugin_dialog import QtPluginDialog

    assert QtPluginDialog is not None


def test_plugin_manager(make_napari_viewer):
    """Test that the plugin manager is accessible from the viewer"""
    viewer = make_napari_viewer()
    assert _plugins._plugin_manager_dialog_avail()

    # Check plugin install action is visible
    plugin_install_action = viewer.window.plugins_menu.findAction(
        'napari.window.plugins.plugin_install_dialog',
    )
    assert plugin_install_action.isVisible()


def test_no_plugin_manager(monkeypatch, make_napari_viewer):
    """Test that the plugin manager menu item is hidden when not installed."""

    def mockreturn(*args):
        return None

    monkeypatch.setattr('importlib.util.find_spec', mockreturn)
    # We need to reload `_plugins` for the monkeypatching to work
    importlib.reload(_plugins)

    assert not _plugins._plugin_manager_dialog_avail()
    # Check plugin install action is not visible
    viewer = make_napari_viewer()
    plugin_install_action = viewer.window.plugins_menu.findAction(
        'napari.window.plugins.plugin_install_dialog',
    )
    assert not plugin_install_action.isVisible()


def test_plugins_menu_sorted(
    mock_pm,  # noqa: F811
    mock_app_model,
    tmp_plugin: DynamicPlugin,
):
    from napari._app_model import get_app_model
    from napari.plugins import _initialize_plugins

    # we make sure 'plugin-b' is registered first
    tmp_plugin2 = tmp_plugin.spawn(
        name='plugin-b', plugin_manager=mock_pm, register=True
    )
    tmp_plugin1 = tmp_plugin.spawn(
        name='plugin-a', plugin_manager=mock_pm, register=True
    )

    @tmp_plugin1.contribute.widget(display_name='Widget 1')
    def widget1(): ...

    @tmp_plugin1.contribute.widget(display_name='Widget 2')
    def widget2(): ...

    @tmp_plugin2.contribute.widget(display_name='Widget 1')
    def widget2_1(): ...

    @tmp_plugin2.contribute.widget(display_name='Widget 2')
    def widget2_2(): ...

    _initialize_plugins()
    plugins_menu = list(get_app_model().menus.get_menu('napari/plugins'))
    submenus = [item for item in plugins_menu if isinstance(item, SubmenuItem)]
    assert len(submenus) == 2
    assert submenus[0].title == 'plugin-a'
    assert submenus[1].title == 'plugin-b'
